import torch
import torch.nn.functional as F
from torch import nn
from torch.nn import Parameter
from torch_geometric.nn import max_pool_x
from torch_geometric.nn.inits import uniform
from torch_scatter import scatter_mean

from deeprank2.utils.community_pooling import community_pooling, get_preloaded_cluster

# ruff: noqa: ANN001, ANN201


class SGraphAttentionLayer(torch.nn.Module):
    """SGraphAttentionLayer.

    This is a new layer that is similar to the graph attention network but simpler
    z_i =  1 / Ni \\Sum_j a_ij * [x_i || x_j] * W + b_i
    || is the concatenation operator: [1,2,3] || [4,5,6] = [1,2,3,4,5,6]
    Ni is the number of neighbor of node i
    \\Sum_j runs over the neighbors of node i
    a_ij is the edge attribute between node i and j
    Args:
        in_channels (int): Size of each input sample.
        out_channels (int): Size of each output sample.
        bias (bool, optional): If set to :obj:`False`, the layer will not learn
            an additive bias. Defaults to True.
    """  # noqa: D301

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        bias: bool = True,
        undirected: bool = True,
    ):
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.undirected = undirected

        self.weight = Parameter(torch.Tensor(2 * in_channels, out_channels))

        if bias:
            self.bias = Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        size = 2 * self.in_channels
        uniform(size, self.weight)
        uniform(size, self.bias)

    def forward(self, x, edge_index, edge_attr):
        row, col = edge_index
        num_node = len(x)
        edge_attr = edge_attr.unsqueeze(-1) if edge_attr.dim() == 1 else edge_attr

        # create edge feature by concatenating node feature
        alpha = torch.cat([x[row], x[col]], dim=-1)

        # multiply the edge features with the fliter
        alpha = torch.mm(alpha, self.weight)

        # multiply each edge features with the corresponding dist
        alpha = edge_attr * alpha

        # scatter the resulting edge feature to get node features
        out = torch.zeros(num_node, self.out_channels).to(alpha.device)
        out = scatter_mean(alpha, row, dim=0, out=out)

        # if the graph is undirected and (i,j) and (j,i) are both in
        # the edge_index then we do not need to have that second line
        # or we count everything twice
        if not self.undirected:
            out = scatter_mean(alpha, col, dim=0, out=out)

        # add the bias
        if self.bias is not None:
            out = out + self.bias

        return out

    def __repr__(self):
        return f"{self.__class__.__name__}({self.in_channels}, {self.out_channels})"


class SGAT(torch.nn.Module):  # noqa:D101
    def __init__(
        self,
        input_shape,
        output_shape=1,
        input_shape_edge=None,  # noqa: ARG002
    ):
        super().__init__()

        self.conv1 = SGraphAttentionLayer(input_shape, 16)
        self.conv2 = SGraphAttentionLayer(16, 32)

        self.fc1 = torch.nn.Linear(32, 64)
        self.fc2 = torch.nn.Linear(64, output_shape)

        self.clustering = "mcl"

    def forward(self, data):
        act = nn.Tanhshrink()
        act = F.relu

        # first conv block
        data.x = act(self.conv1(data.x, data.edge_index, data.edge_attr))
        cluster = get_preloaded_cluster(data.cluster0, data.batch)
        data = community_pooling(cluster, data)

        # second conv block
        data.x = act(self.conv2(data.x, data.edge_index, data.edge_attr))
        cluster = get_preloaded_cluster(data.cluster1, data.batch)
        x, batch = max_pool_x(cluster, data.x, data.batch)

        # FC
        x = scatter_mean(x, batch, dim=0)
        x = act(self.fc1(x))
        x = self.fc2(x)

        return x  # noqa:RET504 (unnecessary-assign)



import logging
import os
import random
from math import sqrt

import pandas as pd
from matplotlib import pyplot as plt
from sklearn.metrics import roc_auc_score
from torch import argmax, tensor
from torch.nn.functional import cross_entropy
from torch.utils.tensorboard import SummaryWriter

_log = logging.getLogger(__name__)


class OutputExporter:
    """The class implements a general exporter to be called when a neural network generates outputs."""

    def __init__(self, directory_path: str | None = None):
        if directory_path is None:
            directory_path = "./output"
        self._directory_path = directory_path

        if not os.path.exists(self._directory_path):
            os.makedirs(self._directory_path)

    def __enter__(self):
        """Overridable."""
        return self

    def __exit__(self, exception_type, exception, traceback):  # noqa: ANN001
        """Overridable."""

    def process(
        self,
        pass_name: str,
        epoch_number: int,
        entry_names: list[str],
        output_values: list,
        target_values: list,
        loss: float,
    ) -> None:
        """The entry_names, output_values, target_values MUST have the same length."""

    def is_compatible_with(
        self,
        output_data_shape: int,  # noqa: ARG002
        target_data_shape: int | None = None,  # noqa: ARG002
    ) -> bool:
        """True if this exporter can work with the given data shapes."""
        return True


class OutputExporterCollection:
    """It allows a series of output exporters to be used at the same time."""

    def __init__(self, *args: list[OutputExporter]):
        self._output_exporters = args

    def __enter__(self):
        for output_exporter in self._output_exporters:
            output_exporter.__enter__()

        return self

    def __exit__(self, exception_type, exception, traceback):  # noqa: ANN001
        for output_exporter in self._output_exporters:
            output_exporter.__exit__(exception_type, exception, traceback)

    def process(
        self,
        pass_name: str,
        epoch_number: int,
        entry_names: list[str],
        output_values: list,
        target_values: list,
        loss: float,
    ) -> None:
        for output_exporter in self._output_exporters:
            output_exporter.process(
                pass_name,
                epoch_number,
                entry_names,
                output_values,
                target_values,
                loss,
            )

    def __iter__(self):
        return iter(self._output_exporters)


class TensorboardBinaryClassificationExporter(OutputExporter):
    """Exporter for tensorboard, works for binary classification only.

    Currently outputs to tensorboard:
    - Mathews Correlation Coefficient (MCC)
    - Accuracy
    - ROC area under the curve
    Outputs are computed for each epoch.
    """

    def __init__(self, directory_path: str):
        super().__init__(directory_path)
        self._writer = SummaryWriter(log_dir=directory_path)

    def __enter__(self):
        self._writer.__enter__()
        return self

    def __exit__(self, exception_type, exception, traceback):  # noqa: ANN001
        self._writer.__exit__(exception_type, exception, traceback)

    def process(
        self,
        pass_name: str,
        epoch_number: int,
        entry_names: list[str],
        output_values: list,
        target_values: list,
        loss: float,  # noqa: ARG002
    ) -> None:
        """Write to tensorboard."""
        ce_loss = cross_entropy(tensor(output_values), tensor(target_values)).item()
        self._writer.add_scalar(
            f"{pass_name} cross entropy loss",
            ce_loss,
            epoch_number,
        )

        probabilities = []
        fp, fn, tp, tn = 0, 0, 0, 0
        for entry_index, _ in enumerate(entry_names):
            probability = output_values[entry_index][1]
            probabilities.append(probability)

            prediction_value = argmax(tensor(output_values[entry_index]))
            target_value = target_values[entry_index]

            if prediction_value > 0 and target_value > 0:
                tp += 1

            elif prediction_value <= 0 and target_value <= 0:
                tn += 1

            elif target_value <= 0 < prediction_value:
                fp += 1

            elif prediction_value <= 0 < target_value:
                fn += 1

        mcc_numerator = tn * tp - fp * fn
        if mcc_numerator == 0:
            self._writer.add_scalar(f"{pass_name} MCC", 0.0, epoch_number)
        else:
            mcc_denominator = sqrt((tn + fn) * (fp + tp) * (tn + fp) * (fn + tp))

            if mcc_denominator != 0:
                mcc = mcc_numerator / mcc_denominator
                self._writer.add_scalar(f"{pass_name} MCC", mcc, epoch_number)

        accuracy = (tp + tn) / (tp + tn + fp + fn)
        self._writer.add_scalar(f"{pass_name} accuracy", accuracy, epoch_number)

        # for ROC curves to work, we need both class values in the set
        if len(set(target_values)) == 2:  # noqa:PLR2004
            roc_auc = roc_auc_score(target_values, probabilities)
            self._writer.add_scalar(f"{pass_name} ROC AUC", roc_auc, epoch_number)

    def is_compatible_with(
        self,
        output_data_shape: int,
        target_data_shape: int | None = None,
    ) -> bool:
        """For regression, target data is needed and output data must be a list of two-dimensional values."""
        return output_data_shape == 2 and target_data_shape == 1  # noqa:PLR2004


class ScatterPlotExporter(OutputExporter):
    """An output exporter that can make scatter plots, containing every single data point.

    On the X-axis: targets values
    On the Y-axis: output values

    Args:
        directory_path (str): Where to store the plots.
        epoch_interval (int, optional): How often to make a plot, 5 means: every 5 epochs. Defaults to 1.
    """

    def __init__(self, directory_path: str, epoch_interval: int = 1):
        super().__init__(directory_path)
        self._epoch_interval = epoch_interval

    def __enter__(self):
        self._plot_data = {}
        return self

    def __exit__(self, exception_type, exception, traceback):  # noqa: ANN001
        self._plot_data.clear()

    def get_filename(self, epoch_number: int) -> str:
        """Returns the filename for the table."""
        return os.path.join(self._directory_path, f"scatter-{epoch_number}.png")

    @staticmethod
    def _get_color(pass_name: str) -> str:
        pass_name = pass_name.lower().strip()
        if pass_name in ("train", "training"):
            return "blue"
        if pass_name in ("eval", "valid", "validation"):
            return "red"
        if pass_name == ("test", "testing"):
            return "green"
        return random.choice(["yellow", "cyan", "magenta"])

    @staticmethod
    def _plot(
        epoch_number: int,
        data: dict[str, tuple[list[float], list[float]]],
        png_path: str,
    ) -> None:
        plt.title(f"Epoch {epoch_number}")

        for pass_name, (truth_values, prediction_values) in data.items():
            plt.scatter(
                truth_values,
                prediction_values,
                color=ScatterPlotExporter._get_color(pass_name),
                label=pass_name,
            )

        plt.xlabel("truth")
        plt.ylabel("prediction")

        plt.legend()
        plt.savefig(png_path)
        plt.close()

    def process(
        self,
        pass_name: str,
        epoch_number: int,
        entry_names: list[str],  # noqa: ARG002
        output_values: list,
        target_values: list,
        loss: float,  # noqa: ARG002
    ) -> None:
        """Make the plot, if the epoch matches with the interval."""
        if epoch_number % self._epoch_interval == 0:
            if epoch_number not in self._plot_data:
                self._plot_data[epoch_number] = {}

            self._plot_data[epoch_number][pass_name] = (target_values, output_values)

            path = self.get_filename(epoch_number)
            self._plot(epoch_number, self._plot_data[epoch_number], path)

    def is_compatible_with(
        self,
        output_data_shape: int,
        target_data_shape: int | None = None,
    ) -> bool:
        """For regression, target data is needed and output data must be a list of one-dimensional values."""
        return output_data_shape == 1 and target_data_shape == 1


class HDF5OutputExporter(OutputExporter):
    """An output exporter that saves every single data point in an hdf5 file.

    It is the most general output exporter implemented, and the information
    contained in the hdf5 file generated allows the user to compute any kind
    of metrics, for both classification and regression.
    Results saved are:
    - phase (train/valid/test)
    - epoch
    - entry name
    - output value/s
    - target value
    - loss per epoch
    The user can then read the content of the hdf5 file into a Pandas dataframe.
    """

    def __init__(self, directory_path: str):
        self.phase = None
        super().__init__(directory_path)

    def __enter__(self):
        self.d = {
            "phase": [],
            "epoch": [],
            "entry": [],
            "output": [],
            "target": [],
            "loss": [],
        }
        self.df = pd.DataFrame(data=self.d)

        return self

    def __exit__(self, exception_type, exception, traceback):  # noqa: ANN001
        if self.phase is not None:
            if self.phase == "validation":
                self.phase = "training"

            self.df.to_hdf(
                os.path.join(self._directory_path, "output_exporter.hdf5"),
                key=self.phase,
                mode="a",
            )

    def process(
        self,
        pass_name: str,
        epoch_number: int,
        entry_names: list[str],
        output_values: list,
        target_values: list,
        loss: float,
    ) -> None:
        self.phase = pass_name
        pass_name = [pass_name] * len(output_values)
        loss = [loss] * len(output_values)
        epoch_number = [epoch_number] * len(output_values)

        d_epoch = {
            "phase": pass_name,
            "epoch": epoch_number,
            "entry": entry_names,
            "output": output_values,
            "target": target_values,
            "loss": loss,
        }
        df_epoch = pd.DataFrame(data=d_epoch)

        self.df = pd.concat([self.df, df_epoch])
        self.df = self.df.reset_index(drop=True)



import importlib
import logging
import os
import pickle
import pkgutil
import re
import warnings
from collections.abc import Iterator
from dataclasses import MISSING, dataclass, field, fields
from functools import partial
from glob import glob
from multiprocessing import Pool
from random import randrange
from types import ModuleType
from typing import Literal

import h5py
import numpy as np
import pdb2sql

import deeprank2.features
from deeprank2.domain.aminoacidlist import convert_aa_nomenclature
from deeprank2.features import components, conservation, contact
from deeprank2.molstruct.aminoacid import AminoAcid
from deeprank2.molstruct.residue import Residue, SingleResidueVariant
from deeprank2.molstruct.structure import PDBStructure
from deeprank2.utils.buildgraph import get_contact_atoms, get_structure, get_surrounding_residues
from deeprank2.utils.graph import Graph
from deeprank2.utils.grid import Augmentation, GridSettings, MapMethod
from deeprank2.utils.parsing.pssm import parse_pssm

_log = logging.getLogger(__name__)

VALID_RESOLUTIONS = ["atom", "residue"]


@dataclass(repr=False, kw_only=True)
class Query:
    """Represents one entity of interest: a single residue variant (SRV) or a protein-protein interface (PPI).

    :class:`Query` objects are used to generate graphs from structures, and they should be created before any model is loaded.
    They can have target values associated with them, which will be stored with the resulting graph.

    Args:
        pdb_path (str): the path to the PDB file to query.
        resolution (Literal['residue', 'atom']): sets whether each node is a residue or atom.
        chain_ids (list[str] | str): the chain identifier(s) of the variant residue or interacting interfaces.
            Note that this does not limit the structure to residues from this/these chain(s).
        pssm_paths (dict[str, str]): the name of the chain(s) (key) and path to the pssm file(s) (value).
        targets (dict[str, float]) = Name(s) (key) and target value(s) (value) associated with this query.
        influence_radius (float | None): all residues within this radius from the variant residue or interacting interfaces
            will be included in the graph, irrespective of the chain they are on.
        max_edge_length (float | None): the maximum distance between two nodes to generate an edge connecting them.
        suppress_pssm_errors (bool): Whether or not to suppress the error raised if the .pssm files do not
            match the .pdb files. If True, a warning is returned instead.
    """

    pdb_path: str
    resolution: Literal["residue", "atom"]
    chain_ids: list[str] | str
    pssm_paths: dict[str, str] = field(default_factory=dict)
    targets: dict[str, float] = field(default_factory=dict)
    influence_radius: float | None = None
    max_edge_length: float | None = None
    suppress_pssm_errors: bool = False

    def __post_init__(self):
        self._model_id = os.path.splitext(os.path.basename(self.pdb_path))[0]
        self.variant = None  # not used for PPI, overwritten for SRV

        if self.resolution == "residue":
            self.max_edge_length = 10 if not self.max_edge_length else self.max_edge_length
            self.influence_radius = 10 if not self.influence_radius else self.influence_radius
        elif self.resolution == "atom":
            self.max_edge_length = 4.5 if not self.max_edge_length else self.max_edge_length
            self.influence_radius = 4.5 if not self.influence_radius else self.influence_radius
        else:
            msg = f"Invalid resolution given ({self.resolution}). Must be one of {VALID_RESOLUTIONS}"
            raise ValueError(msg)

        if not isinstance(self.chain_ids, list):
            self.chain_ids = [self.chain_ids]

        # convert None to empty type (e.g. list, dict) for arguments where this is expected
        for f in fields(self):
            value = getattr(self, f.name)
            if value is None and f.default_factory is not MISSING:
                setattr(self, f.name, f.default_factory())

    def _set_graph_targets(self, graph: Graph) -> None:
        """Copy target data from query to graph."""
        for target_name, target_data in self.targets.items():
            graph.targets[target_name] = target_data

    def _load_structure(self) -> PDBStructure:
        """Build PDBStructure objects from pdb and pssm data."""
        pdb = pdb2sql.pdb2sql(self.pdb_path)
        try:
            structure = get_structure(pdb, self.model_id)
        finally:
            pdb._close()  # noqa: SLF001
        # read the pssm
        if self._pssm_required:
            self._load_pssm_data(structure)

        return structure

    def _load_pssm_data(self, structure: PDBStructure) -> None:
        self._check_pssm()
        for chain in structure.chains:
            if chain.id in self.pssm_paths:
                pssm_path = self.pssm_paths[chain.id]
                with open(pssm_path, encoding="utf-8") as f:
                    chain.pssm = parse_pssm(f, chain)

    def _check_pssm(self, verbosity: Literal[0, 1, 2] = 0) -> None:  # noqa: C901
        """Checks whether information stored in pssm file matches the corresponding pdb file.

        Args:
            pdb_path (str): Path to the PDB file.
            pssm_paths (dict[str, str]): The paths to the PSSM files, per chain identifier.
            suppress (bool): Suppress errors and throw warnings instead.
            verbosity (int): Level of verbosity of error/warning. Defaults to 0.
                0 (low): Only state file name where error occurred;
                1 (medium): Also state number of incorrect and missing residues;
                2 (high): Also list the incorrect residues

        Raises:
            ValueError: Raised if info between pdb file and pssm file doesn't match or if no pssms were provided
        """
        if not self.pssm_paths:
            msg = "No pssm paths provided for conservation feature module."
            raise ValueError(msg)

        # load residues from pssm and pdb files
        pssm_file_residues = {}
        for chain, pssm_path in self.pssm_paths.items():
            with open(pssm_path, encoding="utf-8") as f:
                lines = f.readlines()[1:]
            for line in lines:
                pssm_file_residues[chain + line.split()[0].zfill(4)] = convert_aa_nomenclature(line.split()[1], 3)
        pdb_file_residues = {res[0] + str(res[2]).zfill(4): res[1] for res in pdb2sql.pdb2sql(self.pdb_path).get_residues() if res[0] in self.pssm_paths}

        # list errors
        mismatches = []
        missing_entries = []
        for residue in pdb_file_residues:
            try:
                if pdb_file_residues[residue] != pssm_file_residues[residue]:
                    mismatches.append(residue)
            except KeyError:  # noqa: PERF203
                missing_entries.append(residue)

        # generate error message
        if len(mismatches) + len(missing_entries) > 0:
            error_message = f"Amino acids in PSSM files do not match pdb file for {os.path.split(self.pdb_path)[1]}."
            if verbosity:
                if len(mismatches) > 0:
                    error_message = error_message + f"\n\t{len(mismatches)} entries are incorrect."
                    if verbosity == 2:  # noqa: PLR2004
                        error_message = error_message[-1] + f":\n\t{missing_entries}"
                if len(missing_entries) > 0:
                    error_message = error_message + f"\n\t{len(missing_entries)} entries are missing."
                    if verbosity == 2:  # noqa: PLR2004
                        error_message = error_message[-1] + f":\n\t{missing_entries}"

            # raise exception (or warning)
            if not self.suppress_pssm_errors:
                raise ValueError(error_message)
            warnings.warn(error_message)
            _log.warning(error_message)

    @property
    def model_id(self) -> str:
        """The ID of the model, usually a .PDB accession code."""
        return self._model_id

    @model_id.setter
    def model_id(self, value: str) -> None:
        self._model_id = value

    def __repr__(self) -> str:
        return f"{type(self)}({self.get_query_id()})"

    def build(
        self,
        feature_modules: list[str | ModuleType],
    ) -> Graph:
        """Builds the graph from the .PDB structure.

        Args:
            feature_modules (list[str]): the feature modules used to build the graph.
                These must be filenames existing inside `deeprank2.features` subpackage.

        Returns:
            :class:`Graph`: The resulting :class:`Graph` object with all the features and targets.
        """
        if not isinstance(feature_modules, list):
            feature_modules = [feature_modules]
        feature_modules = [importlib.import_module("deeprank2.features." + module) if isinstance(module, str) else module for module in feature_modules]
        self._pssm_required = conservation in feature_modules
        graph = self._build_helper()

        # add target and feature data to the graph
        self._set_graph_targets(graph)
        for feature_module in feature_modules:
            feature_module.add_features(self.pdb_path, graph, self.variant)

        return graph

    def _build_helper(self) -> Graph:
        msg = "Must be defined in child classes."
        raise NotImplementedError(msg)

    def get_query_id(self) -> str:
        msg = "Must be defined in child classes."
        raise NotImplementedError(msg)


@dataclass(kw_only=True)
class SingleResidueVariantQuery(Query):
    """A query that builds a single residue variant graph.

    Args (common for `Query`):
        pdb_path (str): the path to the PDB file to query.
        resolution (Literal['residue', 'atom']): sets whether each node is a residue or atom.
        chain_ids (list[str] | str): the chain identifier of the variant residue (generally a single capital letter).
            Note that this does not limit the structure to residues from this chain.
        pssm_paths (dict[str, str]): the name of the chain(s) (key) and path to the pssm file(s) (value).
        targets (dict[str, float]) = Name(s) (key) and target value(s) (value) associated with this query.
        influence_radius (float | None): all residues within this radius from the variant residue
            will be included in the graph, irrespective of the chain they are on.
        max_edge_length (float | None): the maximum distance between two nodes to generate an edge connecting them.

        suppress_pssm_errors (bool): Whether or not to suppress the error raised if the .pssm files do not
            match the .pdb files. If True, a warning is returned instead.
    SRV specific Args:
        variant_residue_number (int): the residue number of the variant residue.
        insertion_code (str | None): the insertion code of the variant residue.
        wildtype_amino_acid (AminoAcid): the amino acid at above position in the wildtype protein.
        variant_amino_acid (AminoAcid): the amino acid at above position in the variant protein.
        radius (float): all Residues within this radius (in Å) from the variant residue will
            be included in the graph
    """

    variant_residue_number: int
    insertion_code: str | None
    wildtype_amino_acid: AminoAcid
    variant_amino_acid: AminoAcid

    def __post_init__(self):
        super().__post_init__()  # calls __post_init__ of parents

        if len(self.chain_ids) != 1:
            raise ValueError("`chain_ids` must contain exactly 1 chain for `SingleResidueVariantQuery` objects, " + f"but {len(self.chain_ids)} were given.")
        self.variant_chain_id = self.chain_ids[0]

    @property
    def residue_id(self) -> str:
        """String representation of the residue number and insertion code."""
        if self.insertion_code is not None:
            return f"{self.variant_residue_number}{self.insertion_code}"
        return str(self.variant_residue_number)

    def get_query_id(self) -> str:
        """Returns the string representing the complete query ID."""
        return (
            f"{self.resolution}-srv:"
            f"{self.variant_chain_id}:{self.residue_id}:"
            f"{self.wildtype_amino_acid.name}->{self.variant_amino_acid.name}:{self.model_id}"
        )

    def _build_helper(self) -> Graph:
        """Helper function to build a graph for SRV queries.

        Returns:
            :class:`Graph`: The resulting :class:`Graph` object with all the features and targets.
        """
        # load .PDB structure
        structure = self._load_structure()

        # find the variant residue and its surroundings
        variant_residue: Residue = None
        for residue in structure.get_chain(self.variant_chain_id).residues:
            if residue.number == self.variant_residue_number and residue.insertion_code == self.insertion_code:
                variant_residue = residue
                break
        if variant_residue is None:
            msg = f"Residue not found in {self.pdb_path}: {self.variant_chain_id} {self.residue_id}"
            raise ValueError(msg)
        self.variant = SingleResidueVariant(variant_residue, self.variant_amino_acid)
        residues = get_surrounding_residues(
            structure,
            variant_residue,
            self.influence_radius,
        )

        # build the graph
        if self.resolution == "residue":
            graph = Graph.build_graph(
                residues,
                self.get_query_id(),
                self.max_edge_length,
            )
        elif self.resolution == "atom":
            residues.append(variant_residue)
            atoms = set()
            for residue in residues:
                if residue.amino_acid is not None:
                    for atom in residue.atoms:
                        atoms.add(atom)
            atoms = list(atoms)

            graph = Graph.build_graph(atoms, self.get_query_id(), self.max_edge_length)

        else:
            msg = f"No function exists to build graphs with resolution of {self.resolution}."
            raise NotImplementedError(msg)
        graph.center = variant_residue.get_center()

        return graph


@dataclass(kw_only=True)
class ProteinProteinInterfaceQuery(Query):
    """A query that builds a protein-protein interface graph.

    Args:
        pdb_path (str): the path to the PDB file to query.
        resolution (Literal['residue', 'atom']): sets whether each node is a residue or atom.
        chain_ids (list[str] | str): the chain identifiers of the interacting interfaces (generally a single capital letter each).
            Note that this does not limit the structure to residues from these chains.
        pssm_paths (dict[str, str]): the name of the chain(s) (key) and path to the pssm file(s) (value).
        targets (dict[str, float]) = Name(s) (key) and target value(s) (value) associated with this query.
        influence_radius (float | None): all residues within this radius from the interacting interface
            will be included in the graph, irrespective of the chain they are on.
        max_edge_length (float | None): the maximum distance between two nodes to generate an edge connecting them.
        suppress_pssm_errors (bool): Whether or not to suppress the error raised if the .pssm files do not
            match the .pdb files. If True, a warning is returned instead.
    """

    def __post_init__(self):
        super().__post_init__()

        if len(self.chain_ids) != 2:  # noqa: PLR2004
            raise ValueError(
                "`chain_ids` must contain exactly 2 chains for `ProteinProteinInterfaceQuery` objects, " + f"but {len(self.chain_ids)} was/were given.",
            )

    def get_query_id(self) -> str:
        """Returns the string representing the complete query ID."""
        return (
            f"{self.resolution}-ppi:"  # resolution and query type (ppi for protein protein interface)
            f"{self.chain_ids[0]}-{self.chain_ids[1]}:{self.model_id}"
        )

    def _build_helper(self) -> Graph:
        """Helper function to build a graph for PPI queries.

        Returns:
            :class:`Graph`: The resulting :class:`Graph` object with all the features and targets.
        """
        # find the atoms near the contact interface
        contact_atoms = get_contact_atoms(
            self.pdb_path,
            self.chain_ids,
            self.influence_radius,
        )
        if len(contact_atoms) == 0:
            msg = "No contact atoms found"
            raise ValueError(msg)

        # build the graph
        if self.resolution == "atom":
            graph = Graph.build_graph(
                contact_atoms,
                self.get_query_id(),
                self.max_edge_length,
            )
        elif self.resolution == "residue":
            residues_selected = list({atom.residue for atom in contact_atoms})
            graph = Graph.build_graph(
                residues_selected,
                self.get_query_id(),
                self.max_edge_length,
            )

        graph.center = np.mean([atom.position for atom in contact_atoms], axis=0)
        structure = contact_atoms[0].residue.chain.model
        if self._pssm_required:
            self._load_pssm_data(structure)

        return graph


class QueryCollection:
    """Represents the collection of data queries that will be processed.

    The class attributes are set either while adding queries to the collection (`_queries`
    and `_ids_count`), or when processing the collection (other attributes).

    Attributes:
        _queries (list[:class:`Query`]): The `Query` objects in the collection.
        _ids_count (dict[str, int]): The original `query_id` and the repeat number for this id.
            This is used to rename the `query_id` to ensure that there are no duplicate ids.
        _prefix, _cpu_count, _grid_settings, etc.: See docstring for `QueryCollection.process`.

    Notes:
        Queries can be saved as a dictionary to easily navigate through their data,
        using `QueryCollection.export_dict()`.
    """

    def __init__(self):
        self._queries: list[Query] = []
        self._ids_count: dict[str, int] = {}
        self._prefix: str | None = None
        self._cpu_count: int | None = None
        self._grid_settings: GridSettings | None = None
        self._grid_map_method: MapMethod | None = None
        self._grid_augmentation_count: int = 0

    def add(
        self,
        query: Query,
        verbose: bool = False,
        warn_duplicate: bool = True,
    ) -> None:
        """Add a new query to the collection.

        Args:
            query(:class:`Query`): The `Query` to add to the collection.
            verbose(bool): For logging query IDs added. Defaults to `False`.
            warn_duplicate (bool): Log a warning before renaming if a duplicate query is identified. Defaults to `True`.
        """
        query_id = query.get_query_id()
        if verbose:
            _log.info(f"Adding query with ID {query_id}.")

        if query_id not in self._ids_count:
            self._ids_count[query_id] = 1
        else:
            self._ids_count[query_id] += 1
            new_id = query.model_id + "_" + str(self._ids_count[query_id])
            query.model_id = new_id
            if warn_duplicate:
                _log.warning(f"Query with ID {query_id} has already been added to the collection. Renaming it as {query.get_query_id()}")

        self._queries.append(query)

    def export_dict(self, dataset_path: str) -> None:
        """Exports the colection of all queries to a dictionary file.

        Args:
            dataset_path (str): The path where to save the list of queries.
        """
        with open(dataset_path, "wb") as pkl_file:
            pickle.dump(self, pkl_file)

    @property
    def queries(self) -> list[Query]:
        """The list of queries added to the collection."""
        return self._queries

    def __contains__(self, query: Query) -> bool:
        return query in self._queries

    def __iter__(self) -> Iterator[Query]:
        return iter(self._queries)

    def __len__(self) -> int:
        return len(self._queries)

    def _process_one_query(self, query: Query) -> None:
        """Only one process may access an hdf5 file at a time."""
        try:
            output_path = f"{self._prefix}-{os.getpid()}.hdf5"
            graph = query.build(self._feature_modules)
            graph.write_to_hdf5(output_path)

            if self._grid_settings is not None and self._grid_map_method is not None:
                graph.write_as_grid_to_hdf5(
                    output_path,
                    self._grid_settings,
                    self._grid_map_method,
                )
                for _ in range(self._grid_augmentation_count):
                    # repeat with random augmentation
                    axis, angle = pdb2sql.transform.get_rot_axis_angle(randrange(100))
                    augmentation = Augmentation(axis, angle)
                    graph.write_as_grid_to_hdf5(
                        output_path,
                        self._grid_settings,
                        self._grid_map_method,
                        augmentation,
                    )

        except (ValueError, AttributeError, KeyError, TimeoutError) as e:
            _log.warning(
                f"\nGraph/Query with ID {query.get_query_id()} ran into an Exception ({e.__class__.__name__}: {e}),"
                " and it has not been written to the hdf5 file. More details below:",
            )
            _log.exception(e)

    def process(
        self,
        prefix: str = "processed-queries",
        feature_modules: list[ModuleType, str] | ModuleType | str | None = None,
        cpu_count: int | None = None,
        combine_output: bool = True,
        grid_settings: GridSettings | None = None,
        grid_map_method: MapMethod | None = None,
        grid_augmentation_count: int = 0,
    ) -> list[str]:
        """Render queries into graphs (and optionally grids).

        Args:
            prefix (str | None, optional): Prefix for naming the output files. Defaults to "processed-queries".
            feature_modules (list[ModuleType] | list[str] | Literal ['all'], optional): Feature module or list of feature modules
                used to generate features (given as string or as an imported module).
                Each module must implement the :py:func:`add_features` function, and all feature modules must exist inside `deeprank2.features` folder.
                If set to 'all', all available modules in `deeprank2.features` are used to generate the features.
                Defaults to the two primary feature modules `deeprank2.features.components` and `deeprank2.features.contact`.
            cpu_count (int | None, optional): The number of processes to be run in parallel (i.e. number of CPUs used), capped by
                the number of CPUs available to the system.
                Defaults to None, which takes all available cpu cores.
            combine_output (bool, optional):
                If `True` (default): all processes are combined into a single HDF5 file.
                If `False`: separate HDF5 files are created for each process (i.e. for each CPU used).
            grid_settings (:class:`GridSettings` | None, optional): If valid together with `grid_map_method`, the grid data will be stored as well.
                Defaults to None.
            grid_map_method (:class:`MapMethod` | None, optional): If valid together with `grid_settings`, the grid data will be stored as well.
                Defaults to None.
            grid_augmentation_count (int, optional): Number of grid data augmentations (must be >= 0).
                Defaults to 0.

        Returns:
            list[str]: The list of paths of the generated HDF5 files.
        """
        # set defaults
        feature_modules = feature_modules or [components, contact]
        self._prefix = "processed-queries" if not prefix else re.sub(".hdf5$", "", prefix)  # scrape extension if present

        max_cpus = os.cpu_count()
        self._cpu_count = max_cpus if cpu_count is None else min(cpu_count, max_cpus)
        if cpu_count and self._cpu_count < cpu_count:
            _log.warning(f"\nTried to set {cpu_count} CPUs, but only {max_cpus} are present in the system.")
        _log.info(f"\nNumber of CPUs for processing the queries set to: {self._cpu_count}.")

        self._feature_modules = self._set_feature_modules(feature_modules)
        _log.info(f"\nSelected feature modules: {self._feature_modules}.")

        self._grid_settings = grid_settings
        self._grid_map_method = grid_map_method

        if grid_augmentation_count < 0:
            msg = f"`grid_augmentation_count` cannot be negative, but was given as {grid_augmentation_count}"
            raise ValueError(msg)
        self._grid_augmentation_count = grid_augmentation_count

        _log.info(f"Creating pool function to process {len(self)} queries...")
        pool_function = partial(self._process_one_query)
        with Pool(self._cpu_count) as pool:
            _log.info("Starting pooling...\n")
            pool.map(pool_function, self.queries)

        output_paths = glob(f"{prefix}-*.hdf5")
        if combine_output:
            for output_path in output_paths:
                with h5py.File(f"{prefix}.hdf5", "a") as f_dest, h5py.File(output_path, "r") as f_src:
                    for key, value in f_src.items():
                        _log.debug(f"copy {key} from {output_path} to {prefix}.hdf5")
                        f_src.copy(value, f_dest)
                os.remove(output_path)
            return glob(f"{prefix}.hdf5")

        return output_paths

    def _set_feature_modules(self, feature_modules: list[ModuleType, str] | ModuleType | str) -> list[str]:
        """Convert `feature_modules` to list[str] irrespective of input type.

        Raises:
            TypeError: if an invalid input type is passed.
        """
        if feature_modules == "all":
            return [modname for _, modname, _ in pkgutil.iter_modules(deeprank2.features.__path__)]
        if isinstance(feature_modules, ModuleType):
            return [os.path.basename(feature_modules.__file__)[:-3]]
        if isinstance(feature_modules, str):
            return [re.sub(".py$", "", feature_modules)]  # scrape extension if present
        if isinstance(feature_modules, list):
            invalid_inputs = [type(el) for el in feature_modules if not isinstance(el, str | ModuleType)]
            if invalid_inputs:
                msg = f"`feature_modules` contains invalid input ({invalid_inputs}). Only `str` and `ModuleType` are accepted."
                raise TypeError(msg)
            return [
                re.sub(".py$", "", m) if isinstance(m, str) else os.path.basename(m.__file__)[:-3]  # for ModuleTypes
                for m in feature_modules
            ]
        msg = f"`feature_modules` has received an invalid input type: {type(feature_modules)}. Only `str` and `ModuleType` are accepted."
        raise TypeError(msg)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from typing_extensions import Self

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from deeprank2.molstruct.aminoacid import AminoAcid
    from deeprank2.molstruct.atom import Atom
    from deeprank2.molstruct.structure import Chain
    from deeprank2.utils.pssmdata import PssmRow


class Residue:
    """One protein residue in a `PDBStructure`."""

    def __init__(
        self,
        chain: Chain,
        number: int,
        amino_acid: AminoAcid | None = None,
        insertion_code: str | None = None,
    ):
        """One protein residue in a `PDBStructure`.

        A `Residue` is the basic building block of proteins and protein complex, here represented by `PDBStructures`.
        Each `Residue` is of a certain `AminoAcid` type and consists of multiple `Atom`s.

        Args:
            chain (:class:`Chain`): The chain that this residue belongs to.
            number (int): the residue number
            amino_acid (:class:`AminoAcid`, optional): The residue's amino acid (if it's part of a protein).
                Defaults to None.
            insertion_code (str, optional): The pdb insertion code, if any. Defaults to None.
        """
        self._chain = chain
        self._number = number
        self._amino_acid = amino_acid
        self._insertion_code = insertion_code
        self._atoms = []

    def __eq__(self, other: Self) -> bool:
        if isinstance(other, Residue):
            return self._chain == other._chain and self._number == other._number and self._insertion_code == other._insertion_code
        return NotImplemented

    def __hash__(self) -> hash:
        return hash((self._number, self._insertion_code))

    def get_pssm(self) -> PssmRow:
        """Load pssm info linked to the residue."""
        pssm = self._chain.pssm
        if pssm is None:
            msg = f"No pssm file found for Chain {self._chain}."
            raise FileNotFoundError(msg)
        return pssm[self]

    @property
    def number(self) -> int:
        return self._number

    @property
    def chain(self) -> Chain:
        return self._chain

    @property
    def amino_acid(self) -> AminoAcid:
        return self._amino_acid

    @property
    def atoms(self) -> list[Atom]:
        return self._atoms

    @property
    def number_string(self) -> str:
        """Contains both the number and the insertion code (if any)."""
        if self._insertion_code is not None:
            return f"{self._number}{self._insertion_code}"
        return str(self._number)

    @property
    def insertion_code(self) -> str:
        return self._insertion_code

    def add_atom(self, atom: Atom) -> None:
        self._atoms.append(atom)

    def __repr__(self) -> str:
        return f"{self._chain} {self.number_string}"

    @property
    def position(self) -> np.array:
        return self.get_center()

    def get_center(self) -> NDArray:
        """Find the center position of a `Residue`.

        Center position is found as follows:
        1. find beta carbon
        2. if no beta carbon is found: find alpha carbon
        3. if no alpha carbon is found: take the mean of the atom positions
        """
        betas = [atom for atom in self.atoms if atom.name == "CB"]
        if len(betas) > 0:
            return betas[0].position

        alphas = [atom for atom in self.atoms if atom.name == "CA"]
        if len(alphas) > 0:
            return alphas[0].position

        if len(self.atoms) == 0:
            msg = f"Cannot get the center position from {self}, because it has no atoms"
            raise ValueError(msg)

        return np.mean([atom.position for atom in self.atoms], axis=0)


class SingleResidueVariant:
    """A single residue mutation of a PDBStrcture.

    Args:
        residue (Residue): the `Residue` object from the PDBStructure that is mutated.
        variant_amino_acid (AminoAcid): the amino acid that the `Residue` is mutated into.
    """

    def __init__(self, residue: Residue, variant_amino_acid: AminoAcid):
        self._residue = residue
        self._variant_amino_acid = variant_amino_acid

    @property
    def residue(self) -> Residue:
        return self._residue

    @property
    def variant_amino_acid(self) -> AminoAcid:
        return self._variant_amino_acid

    @property
    def wildtype_amino_acid(self) -> AminoAcid:
        return self._residue.amino_acid



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

import numpy as np
from typing_extensions import Self

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from deeprank2.molstruct.residue import Residue


class AtomicElement(Enum):
    """One-hot encoding of the atomic element (or atom type)."""

    C = 1
    O = 2  # noqa: E741
    N = 3
    S = 4
    P = 5
    H = 6

    @property
    def onehot(self) -> np.array:
        value = np.zeros(max(el.value for el in AtomicElement))
        value[self.value - 1] = 1.0
        return value


class Atom:
    """One atom in a PDBStructure.

    Args:
        residue (:class:`Residue`): The residue that this atom belongs to.
        name (str): Pdb atom name.
        element (:class:`AtomicElement`): The chemical element.
        position (np.array): Pdb position xyz of this atom.
        occupancy (float): Pdb occupancy value.
            This represents the proportion of structures where the atom is detected at a given position.
            Sometimes a single atom can be detected at multiple positions. In that case separate structures exist where sum(occupancy) == 1.
            Note that only the highest occupancy atom is used by deeprank2 (see tools.pdb._add_atom_to_residue).
    """

    def __init__(
        self,
        residue: Residue,
        name: str,
        element: AtomicElement,
        position: NDArray,
        occupancy: float,
    ):
        self._residue = residue
        self._name = name
        self._element = element
        self._position = position
        self._occupancy = occupancy

    def __eq__(self, other: Self) -> bool:
        if isinstance(other, Atom):
            return self._residue == other._residue and self._name == other._name
        return NotImplemented

    def __hash__(self) -> hash:
        return hash((tuple(self._position), self._element, self._name))

    def __repr__(self) -> str:
        return f"{self._residue} {self._name}"

    def change_altloc(self, alternative_atom: Atom) -> None:
        """Replace the atom's location by another atom's location."""
        self._position = alternative_atom.position
        self._occupancy = alternative_atom.occupancy

    @property
    def name(self) -> str:
        return self._name

    @property
    def element(self) -> AtomicElement:
        return self._element

    @property
    def occupancy(self) -> float:
        return self._occupancy

    @property
    def position(self) -> NDArray:
        return self._position

    @property
    def residue(self) -> Residue:
        return self._residue



import re


class ResidueClassCriterium:  # noqa: D101
    def __init__(
        self,
        class_name: str,
        amino_acid_names: str | list[str],
        present_atom_names: list[str],
        absent_atom_names: list[str],
    ):
        self.class_name = class_name

        self.amino_acid_names = amino_acid_names

        self.present_atom_names = present_atom_names
        self.absent_atom_names = absent_atom_names

    def matches(self, amino_acid_name: str, atom_names: list[str]) -> bool:
        # check the amino acid name
        if self.amino_acid_names != "all" and not any(amino_acid_name == crit_amino_acid_name for crit_amino_acid_name in self.amino_acid_names):
            return False

        # check the atom names that should be absent
        if any(atom_name in self.absent_atom_names for atom_name in atom_names):
            return False

        # check the atom names that should be present
        if not all(atom_name in atom_names for atom_name in self.present_atom_names):
            return False

        # all checks passed
        return True


class ResidueClassParser:  # noqa: D101
    _RESIDUE_CLASS_PATTERN = re.compile(r"([A-Z]{3,4}) *\: *name *\= *(all|[A-Z]{3})")
    _RESIDUE_ATOMS_PATTERN = re.compile(r"(present|absent)\(([A-Z0-9\, ]+)\)")

    @staticmethod
    def parse(file_: str) -> list[ResidueClassCriterium]:
        result = []
        for line in file_:
            match = ResidueClassParser._RESIDUE_CLASS_PATTERN.match(line)
            if not match:
                msg = f"Unparsable line: '{line}'"
                raise ValueError(msg)

            class_name = match.group(1)
            amino_acid_names = ResidueClassParser._parse_amino_acids(match.group(2))

            present_atom_names = []
            absent_atom_names = []
            for match in ResidueClassParser._RESIDUE_ATOMS_PATTERN.finditer(line[match.end() :]):  # noqa: B020
                atom_names = [name.strip() for name in match.group(2).split(",")]
                if match.group(1) == "present":
                    present_atom_names.extend(atom_names)

                elif match.group(1) == "absent":
                    absent_atom_names.extend(atom_names)

            result.append(ResidueClassCriterium(class_name, amino_acid_names, present_atom_names, absent_atom_names))
        return result

    @staticmethod
    def _parse_amino_acids(string: str) -> str | list[str]:
        if string.strip() == "all":
            return string.strip()
        return [name.strip() for name in string.split(",")]



from typing import Literal

from deeprank2.molstruct.aminoacid import AminoAcid, Polarity

# All info below sourced from above websites in December 2022 and summarized in deeprank2/domain/aminoacid_summary.xlsx

# Charge is calculated from summing all atoms in the residue (from ./deeprank2/domain/forcefield/protein-allhdg5-5_new.top).
# This results in the expected charge of 0 for all POLAR and NONPOLAR residues, +1 for POSITIVE residues and -1 for NEGATIVE residues.
# Note that SER, THR, and TYR lead to a charge of ~1e-16. A rounding error is assumed in these cases and they are set to 0.

# Sources for Polarity:
#   1) https://www.britannica.com/science/amino-acid/Standard-amino-acids
#   2) https://www.shimadzu.co.jp/aboutus/ms_r/archive/files/AminoAcidTable.pdf
#   3) https://en.wikipedia.org/wiki/Amino_acid
#   3) https://nld.promega.com/resources/tools/amino-acid-chart-amino-acid-structure/
#   5) https://ib.bioninja.com.au/standard-level/topic-2-molecular-biology/24-proteins/amino-acids.html
#   6) print book: "Biology", by Campbell & Reece, 6th ed, ISBN: 0-201-75054-6
# Sources 1, 2, and 6 sources agree on every amino acid and are used below.
# The other sources have some minor discrepancies compared to this and are commented inline.

# Source for size:
#   https://www.shimadzu.co.jp/aboutus/ms_r/archive/files/AminoAcidTable.pdf

# Sources for mass and pI:
#   1) https://www.sigmaaldrich.com/NL/en/technical-documents/technical-article/protein-biology/protein-structural-analysis/amino-acid-reference-chart
#   2) https://www.shimadzu.co.jp/aboutus/ms_r/archive/files/AminoAcidTable.pdf
#   3) https://www.nectagen.com/reference-data/ingredients/amino-acids
# Discrepancies of <0.1 (for either property) between sources are ignored.
# Two instances (LYS and THR) have a larger discrepancy for pI in 1/3 sources; majority rule is implemented (and outlier is indicated in inline comment).

# Sources for hydrogen bond donors and acceptors:
#   1) https://foldit.fandom.com/wiki/Sidechain_Bonding_Gallery
#   2) https://www.imgt.org/IMGTeducation/Aide-memoire/_UK/aminoacids/charge/

# For selenocysteine and pyrrolysine only few sources detailing some of their properties could be found.
# Whatever sources were found (or lack thereof) are indicated in inline comments, but the reliability is much lower than for the canonical amino acids.
# Also, the rest of the package is not expecting these, so they removed from amino_acids at the bottom of this file.

alanine = AminoAcid(
    "Alanine",
    "ALA",
    "A",
    charge=0,
    polarity=Polarity.NONPOLAR,
    size=1,
    mass=71.1,
    pI=6.00,
    hydrogen_bond_donors=0,
    hydrogen_bond_acceptors=0,
    index=0,
)

cysteine = AminoAcid(
    "Cysteine",
    "CYS",
    "C",
    charge=0,
    polarity=Polarity.POLAR,  # source 3: "special case"; source 5: nonpolar
    # polarity of C is generally considered ambiguous: https://chemistry.stackexchange.com/questions/143142/why-is-the-amino-acid-cysteine-classified-as-polar
    size=2,
    mass=103.2,
    pI=5.07,
    hydrogen_bond_donors=0,
    hydrogen_bond_acceptors=0,
    index=1,
)

selenocysteine = AminoAcid(
    "Selenocysteine",
    "SEC",
    "U",
    charge=0,
    polarity=Polarity.POLAR,  # source 3: "special case"
    size=2,  # source: https://en.wikipedia.org/wiki/Selenocysteine
    mass=150.0,  # only from source 3
    pI=5.47,  # only from source 3
    hydrogen_bond_donors=1,  # unconfirmed
    hydrogen_bond_acceptors=2,  # unconfirmed
    index=cysteine.index,
)

aspartate = AminoAcid(
    "Aspartate",
    "ASP",
    "D",
    charge=-1,
    polarity=Polarity.NEGATIVE,
    size=4,
    mass=115.1,
    pI=2.77,
    hydrogen_bond_donors=0,
    hydrogen_bond_acceptors=4,
    index=2,
)

glutamate = AminoAcid(
    "Glutamate",
    "GLU",
    "E",
    charge=-1,
    polarity=Polarity.NEGATIVE,
    size=5,
    mass=129.1,
    pI=3.22,
    hydrogen_bond_donors=0,
    hydrogen_bond_acceptors=4,
    index=3,
)

phenylalanine = AminoAcid(
    "Phenylalanine",
    "PHE",
    "F",
    charge=0,
    polarity=Polarity.NONPOLAR,
    size=7,
    mass=147.2,
    pI=5.48,
    hydrogen_bond_donors=0,
    hydrogen_bond_acceptors=0,
    index=4,
)

glycine = AminoAcid(
    "Glycine",
    "GLY",
    "G",
    charge=0,
    polarity=Polarity.NONPOLAR,  # source 3: "special case"
    size=0,
    mass=57.1,
    pI=5.97,
    hydrogen_bond_donors=0,
    hydrogen_bond_acceptors=0,
    index=5,
)

histidine = AminoAcid(
    "Histidine",
    "HIS",
    "H",
    charge=1,
    polarity=Polarity.POSITIVE,
    size=6,
    mass=137.1,
    pI=7.59,
    hydrogen_bond_donors=1,
    hydrogen_bond_acceptors=1,
    # both position 7 and 10 can serve as either donor or acceptor (depending on tautomer), but any single His will have exactly one donor and one acceptor
    # (see https://foldit.fandom.com/wiki/Histidine)
    index=6,
)

isoleucine = AminoAcid(
    "Isoleucine",
    "ILE",
    "I",
    charge=0,
    polarity=Polarity.NONPOLAR,
    size=4,
    mass=113.2,
    pI=6.02,
    hydrogen_bond_donors=0,
    hydrogen_bond_acceptors=0,
    index=7,
)

lysine = AminoAcid(
    "Lysine",
    "LYS",
    "K",
    charge=1,
    polarity=Polarity.POSITIVE,
    size=5,
    mass=128.2,
    pI=9.74,  # 9.60 in source 3
    hydrogen_bond_donors=3,
    hydrogen_bond_acceptors=0,
    index=8,
)

pyrrolysine = AminoAcid(
    "Pyrrolysine",
    "PYL",
    "O",
    charge=0,  # unconfirmed
    polarity=Polarity.POLAR,  # based on having both H-bond donors and acceptors
    size=13,  # source: https://en.wikipedia.org/wiki/Pyrrolysine
    mass=255.32,  # from source 3
    pI=7.394,  # rough estimate from https://rstudio-pubs-static.s3.amazonaws.com/846259_7a9236df54e6410a972621590ecdcfcb.html
    hydrogen_bond_donors=1,  # unconfirmed
    hydrogen_bond_acceptors=4,  # unconfirmed
    index=lysine.index,
)

leucine = AminoAcid(
    "Leucine",
    "LEU",
    "L",
    charge=0,
    polarity=Polarity.NONPOLAR,
    size=4,
    mass=113.2,
    pI=5.98,
    hydrogen_bond_donors=0,
    hydrogen_bond_acceptors=0,
    index=9,
)

methionine = AminoAcid(
    "Methionine",
    "MET",
    "M",
    charge=0,
    polarity=Polarity.NONPOLAR,
    size=4,
    mass=131.2,
    pI=5.74,
    hydrogen_bond_donors=0,
    hydrogen_bond_acceptors=0,
    index=10,
)

asparagine = AminoAcid(
    "Asparagine",
    "ASN",
    "N",
    charge=0,
    polarity=Polarity.POLAR,
    size=4,
    mass=114.1,
    pI=5.41,
    hydrogen_bond_donors=2,
    hydrogen_bond_acceptors=2,
    index=11,
)

proline = AminoAcid(
    "Proline",
    "PRO",
    "P",
    charge=0,
    polarity=Polarity.NONPOLAR,
    size=3,
    mass=97.1,
    pI=6.30,
    hydrogen_bond_donors=0,
    hydrogen_bond_acceptors=0,
    index=12,
)

glutamine = AminoAcid(
    "Glutamine",
    "GLN",
    "Q",
    charge=0,
    polarity=Polarity.POLAR,
    size=5,
    mass=128.1,
    pI=5.65,
    hydrogen_bond_donors=2,
    hydrogen_bond_acceptors=2,
    index=13,
)

arginine = AminoAcid(
    "Arginine",
    "ARG",
    "R",
    charge=1,
    polarity=Polarity.POSITIVE,
    size=7,
    mass=156.2,
    pI=10.76,
    hydrogen_bond_donors=5,
    hydrogen_bond_acceptors=0,
    index=14,
)

serine = AminoAcid(
    "Serine",
    "SER",
    "S",
    charge=0,
    polarity=Polarity.POLAR,
    size=2,
    mass=87.1,
    pI=5.68,
    hydrogen_bond_donors=1,
    hydrogen_bond_acceptors=2,
    index=15,
)

threonine = AminoAcid(
    "Threonine",
    "THR",
    "T",
    charge=0,
    polarity=Polarity.POLAR,
    size=3,
    mass=101.1,
    pI=5.60,  # 6.16 in source 2
    hydrogen_bond_donors=1,
    hydrogen_bond_acceptors=2,
    index=16,
)

valine = AminoAcid(
    "Valine",
    "VAL",
    "V",
    charge=0,
    polarity=Polarity.NONPOLAR,
    size=3,
    mass=99.1,
    pI=5.96,
    hydrogen_bond_donors=0,
    hydrogen_bond_acceptors=0,
    index=17,
)

tryptophan = AminoAcid(
    "Tryptophan",
    "TRP",
    "W",
    charge=0,
    polarity=Polarity.NONPOLAR,  # source 4: polar
    size=10,
    mass=186.2,
    pI=5.89,
    hydrogen_bond_donors=1,
    hydrogen_bond_acceptors=0,
    index=18,
)

tyrosine = AminoAcid(
    "Tyrosine",
    "TYR",
    "Y",
    charge=-0.0,
    polarity=Polarity.POLAR,  # source 3: nonpolar
    size=8,
    mass=163.2,
    pI=5.66,
    hydrogen_bond_donors=1,
    hydrogen_bond_acceptors=1,
    index=19,
)


# Including selenocysteine and pyrrolysine in the future will require some work to be done on the package.
amino_acids = [
    alanine,
    arginine,
    asparagine,
    aspartate,
    cysteine,
    glutamate,
    glutamine,
    glycine,
    histidine,
    isoleucine,
    leucine,
    lysine,
    methionine,
    phenylalanine,
    proline,
    serine,
    threonine,
    tryptophan,
    tyrosine,
    valine,
    # selenocysteine,
    # pyrrolysine,
]

amino_acids_by_code = {amino_acid.three_letter_code: amino_acid for amino_acid in amino_acids}
amino_acids_by_letter = {amino_acid.one_letter_code: amino_acid for amino_acid in amino_acids}
amino_acids_by_name = {amino_acid.name: amino_acid for amino_acid in amino_acids}


def convert_aa_nomenclature(aa: str, output_format: Literal[0, 1, 3] = 0) -> str:
    """Converts amino acid nomenclatures.

    Conversions are possible between the standard 1-letter codes, 3-letter
    codes, and full names of amino acids.

    Args:
        aa (str): The amino acid to be converted in any of its formats. The
            length of the string is used to determine which format is used.
        output_format (Literal[0, 1, 3], optional): Nomenclature style to return:
            0 (default) returns the full name,
            1 returns the 1-letter code,
            3 returns the 3-letter code.

    Raises:
        ValueError: If aa is not recognized or an invalid output format was given

    Returns:
        str: amino acid in the selected nomenclature system.
    """
    try:
        if len(aa) == 1:
            aa: AminoAcid = next(entry for entry in amino_acids if entry.one_letter_code.lower() == aa.lower())
        elif len(aa) == 3:  # noqa:PLR2004
            aa: AminoAcid = next(entry for entry in amino_acids if entry.three_letter_code.lower() == aa.lower())
        else:
            aa: AminoAcid = next(entry for entry in amino_acids if entry.name.lower() == aa.lower())
    except IndexError as e:
        msg = f"{aa} is not a valid amino acid."
        raise ValueError(msg) from e

    if not output_format:
        return aa.name
    if output_format == 3:  # noqa:PLR2004
        return aa.three_letter_code
    if output_format == 1:
        return aa.one_letter_code
    msg = f"output_format {output_format} not recognized. Must be set to 0 (amino acid name), 1 (one letter code), or 3 (three letter code)."
    raise ValueError(msg)



from typing import TextIO

from deeprank2.domain.aminoacidlist import amino_acids, amino_acids_by_letter
from deeprank2.molstruct.residue import Residue
from deeprank2.molstruct.structure import Chain
from deeprank2.utils.pssmdata import PssmRow, PssmTable


def parse_pssm(file_: TextIO, chain: Chain) -> PssmTable:
    """Read the PSSM data.

    Args:
        file_ (python text file object): The pssm file.
        chain (:class:`Chain`): The chain that the pssm file represents, residues from this chain must match the pssm file.

    Returns:
        PssmTable: The position-specific scoring table, parsed from the pssm file.
    """
    conservation_rows = {}

    # Read the pssm header.
    header = next(file_).split()
    column_indices = {column_name.strip(): index for index, column_name in enumerate(header)}

    for line in file_:
        row = line.split()

        # Read what amino acid the chain is supposed to have at this position.
        amino_acid = amino_acids_by_letter[row[column_indices["pdbresn"]]]

        # Some PDB files have insertion codes, find these to prevent
        # exceptions.
        pdb_residue_number_string = row[column_indices["pdbresi"]]
        if pdb_residue_number_string[-1].isalpha():
            pdb_residue_number = int(pdb_residue_number_string[:-1])
            pdb_insertion_code = pdb_residue_number_string[-1]
        else:
            pdb_residue_number = int(pdb_residue_number_string)
            pdb_insertion_code = None

        # Build the residue, to match with the pssm row
        residue = Residue(chain, pdb_residue_number, amino_acid, pdb_insertion_code)

        # Build the pssm row
        information_content = float(row[column_indices["IC"]])
        conservations = {amino_acid: float(row[column_indices[amino_acid.one_letter_code]]) for amino_acid in amino_acids}

        conservation_rows[residue] = PssmRow(conservations, information_content)

    return PssmTable(conservation_rows)



import logging

import numpy as np

from deeprank2.domain import nodestorage as Nfeat
from deeprank2.molstruct.atom import Atom
from deeprank2.molstruct.residue import Residue, SingleResidueVariant
from deeprank2.utils.graph import Graph
from deeprank2.utils.parsing import atomic_forcefield

_log = logging.getLogger(__name__)


def add_features(  # noqa:D103
    pdb_path: str,  # noqa: ARG001
    graph: Graph,
    single_amino_acid_variant: SingleResidueVariant | None = None,
) -> None:
    for node in graph.nodes:
        if isinstance(node.id, Residue):
            residue = node.id
        elif isinstance(node.id, Atom):
            atom = node.id
            residue = atom.residue

            node.features[Nfeat.ATOMTYPE] = atom.element.onehot
            node.features[Nfeat.PDBOCCUPANCY] = atom.occupancy
            node.features[Nfeat.ATOMCHARGE] = atomic_forcefield.get_charge(atom)
        else:
            msg = f"Unexpected node type: {type(node.id)}"
            raise TypeError(msg)

        node.features[Nfeat.RESTYPE] = residue.amino_acid.onehot
        node.features[Nfeat.RESCHARGE] = residue.amino_acid.charge
        node.features[Nfeat.POLARITY] = residue.amino_acid.polarity.onehot
        node.features[Nfeat.RESSIZE] = residue.amino_acid.size
        node.features[Nfeat.RESMASS] = residue.amino_acid.mass
        node.features[Nfeat.RESPI] = residue.amino_acid.pI
        node.features[Nfeat.HBDONORS] = residue.amino_acid.hydrogen_bond_donors
        node.features[Nfeat.HBACCEPTORS] = residue.amino_acid.hydrogen_bond_acceptors

        if single_amino_acid_variant is not None:
            wildtype = single_amino_acid_variant.wildtype_amino_acid
            variant = single_amino_acid_variant.variant_amino_acid

            if residue == single_amino_acid_variant.residue:
                node.features[Nfeat.VARIANTRES] = variant.onehot
                node.features[Nfeat.DIFFCHARGE] = variant.charge - wildtype.charge
                node.features[Nfeat.DIFFPOLARITY] = variant.polarity.onehot - wildtype.polarity.onehot
                node.features[Nfeat.DIFFSIZE] = variant.size - wildtype.size
                node.features[Nfeat.DIFFMASS] = variant.mass - wildtype.mass
                node.features[Nfeat.DIFFPI] = variant.pI - wildtype.pI
                node.features[Nfeat.DIFFHBDONORS] = variant.hydrogen_bond_donors - wildtype.hydrogen_bond_donors
                node.features[Nfeat.DIFFHBACCEPTORS] = variant.hydrogen_bond_acceptors - wildtype.hydrogen_bond_acceptors
            else:
                node.features[Nfeat.VARIANTRES] = residue.amino_acid.onehot
                node.features[Nfeat.DIFFCHARGE] = 0
                node.features[Nfeat.DIFFPOLARITY] = np.zeros(residue.amino_acid.polarity.onehot.shape)
                node.features[Nfeat.DIFFSIZE] = 0
                node.features[Nfeat.DIFFMASS] = 0
                node.features[Nfeat.DIFFPI] = 0
                node.features[Nfeat.DIFFHBDONORS] = 0
                node.features[Nfeat.DIFFHBACCEPTORS] = 0



"""This module holds the classes that are used when working with a 3D grid."""

import itertools
import logging
from enum import Enum

import h5py
import numpy as np
from numpy.typing import NDArray
from scipy.interpolate import BSpline

from deeprank2.domain import gridstorage

_log = logging.getLogger(__name__)


class MapMethod(Enum):
    """This holds the value of either one of 4 grid mapping methods.

    A mapping method determines how feature point values are divided over the grid points.
    """

    GAUSSIAN = 1
    FAST_GAUSSIAN = 2
    BSP_LINE = 3
    NEAREST_NEIGHBOURS = 4


class Augmentation:
    """A rotation around an axis, to be applied to a feature before mapping it to a grid."""

    def __init__(self, axis: NDArray, angle: float):
        self._axis = axis
        self._angle = angle

    @property
    def axis(self) -> NDArray:
        return self._axis

    @property
    def angle(self) -> float:
        return self._angle


class GridSettings:
    """Objects of this class hold the settings to build a grid.

    The grid is basically a multi-divided 3D cube with
    the following properties:
    - points_counts: the number of points on the x, y, z edges of the cube
    - sizes: x, y, z sizes of the box in Å
    - resolutions: the size in Å of one x, y, z edge subdivision. Also the distance between two points on the edge.
    """

    def __init__(
        self,
        points_counts: list[int],
        sizes: list[float],
    ):
        if len(points_counts) != 3 or len(sizes) != 3:  # noqa:PLR2004
            msg = "Incorrect grid dimensions."
            raise ValueError(msg)

        self._points_counts = points_counts
        self._sizes = sizes

    @property
    def resolutions(self) -> list[float]:
        return [self._sizes[i] / self._points_counts[i] for i in range(3)]

    @property
    def sizes(self) -> list[float]:
        return self._sizes

    @property
    def points_counts(self) -> list[int]:
        return self._points_counts


class Grid:
    """A 3D (volumetric) representation of a `Graph`.

    A Grid contains the following information:

    - coordinates of points
    - names of features
    - feature values on each point.
    """

    def __init__(self, id_: str, center: list[float], settings: GridSettings):
        self.id = id_
        self._center = np.array(center)
        self._settings = settings
        self._set_mesh(self._center, settings)
        self._features = {}

    def _set_mesh(self, center: NDArray, settings: GridSettings) -> None:
        """Builds the grid points."""
        half_size_x = settings.sizes[0] / 2
        half_size_y = settings.sizes[1] / 2
        half_size_z = settings.sizes[2] / 2

        min_x = center[0] - half_size_x
        max_x = min_x + (settings.points_counts[0] - 1.0) * settings.resolutions[0]
        self._xs = np.linspace(min_x, max_x, num=settings.points_counts[0])

        min_y = center[1] - half_size_y
        max_y = min_y + (settings.points_counts[1] - 1.0) * settings.resolutions[1]
        self._ys = np.linspace(min_y, max_y, num=settings.points_counts[1])

        min_z = center[2] - half_size_z
        max_z = min_z + (settings.points_counts[2] - 1.0) * settings.resolutions[2]
        self._zs = np.linspace(min_z, max_z, num=settings.points_counts[2])

        self._ygrid, self._xgrid, self._zgrid = np.meshgrid(self._ys, self._xs, self._zs)

    @property
    def center(self) -> NDArray:
        return self._center

    @property
    def xs(self) -> NDArray:
        return self._xs

    @property
    def xgrid(self) -> NDArray:
        return self._xgrid

    @property
    def ys(self) -> NDArray:
        return self._ys

    @property
    def ygrid(self) -> NDArray:
        return self._ygrid

    @property
    def zs(self) -> NDArray:
        return self._zs

    @property
    def zgrid(self) -> NDArray:
        return self._zgrid

    @property
    def features(self) -> dict[str, NDArray]:
        return self._features

    def add_feature_values(self, feature_name: str, data: NDArray) -> None:
        """Makes sure feature values per grid point get stored.

        This method may be called repeatedly to add on to existing grid point values.
        """
        if feature_name not in self._features:
            self._features[feature_name] = data
        else:
            self._features[feature_name] += data

    def _get_mapped_feature_gaussian(
        self,
        position: NDArray,
        value: float,
    ) -> NDArray:
        beta = 1.0

        fx, fy, fz = position
        distances = np.sqrt((self.xgrid - fx) ** 2 + (self.ygrid - fy) ** 2 + (self.zgrid - fz) ** 2)

        return value * np.exp(-beta * distances)

    def _get_mapped_feature_fast_gaussian(self, position: NDArray, value: float) -> NDArray:
        beta = 1.0
        cutoff = 5.0 * beta

        fx, fy, fz = position
        distances = np.sqrt((self.xgrid - fx) ** 2 + (self.ygrid - fy) ** 2 + (self.zgrid - fz) ** 2)

        data = np.zeros(distances.shape)

        data[distances < cutoff] = value * np.exp(-beta * distances[distances < cutoff])

        return data

    def _get_mapped_feature_bsp_line(
        self,
        position: NDArray,
        value: float,
    ) -> NDArray:
        order = 4

        fx, fy, fz = position
        bsp_data = (
            BSpline((self.xgrid - fx) / self._settings.resolutions[0], order)
            * BSpline((self.ygrid - fy) / self._settings.resolutions[1], order)
            * BSpline((self.zgrid - fz) / self._settings.resolutions[2], order)
        )

        return value * bsp_data

    def _get_mapped_feature_nearest_neighbour(
        self,
        position: NDArray,
        value: float,
    ) -> NDArray:
        fx, _, _ = position
        distances_x = np.abs(self.xs - fx)
        distances_y = np.abs(self.ys - fx)
        distances_z = np.abs(self.zs - fx)

        indices_x = np.argsort(distances_x)[:2]
        indices_y = np.argsort(distances_y)[:2]
        indices_z = np.argsort(distances_z)[:2]

        sorted_x = distances_x[indices_x]
        weights_x = sorted_x / np.sum(sorted_x)

        sorted_y = distances_y[indices_y]
        weights_y = sorted_y / np.sum(sorted_y)

        sorted_z = distances_z[indices_z]
        weights_z = sorted_z / np.sum(sorted_z)

        indices = [indices_x, indices_y, indices_z]
        points = list(itertools.product(*indices))

        weight_products = list(itertools.product(weights_x, weights_y, weights_z))
        weights = [np.sum(p) for p in weight_products]

        neighbour_data = np.zeros((self.xs.shape[0], self.ys.shape[0], self.zs.shape[0]))

        for point_index, point in enumerate(points):
            weight = weights[point_index]

            neighbour_data[point] = weight * value

        return neighbour_data

    def _get_atomic_density_koes(
        self,
        position: NDArray,
        vanderwaals_radius: float,
    ) -> NDArray:
        """Function to map individual atomic density on the grid.

        The formula is equation (1) of the Koes paper
        Protein-Ligand Scoring with Convolutional NN Arxiv:1612.02751v1.

        Returns:
            NDArray: The mapped density.
        """
        distances = np.sqrt(np.square(self.xgrid - position[0]) + np.square(self.ygrid - position[1]) + np.square(self.zgrid - position[2]))

        density_data = np.zeros(distances.shape)

        indices_close = distances < vanderwaals_radius
        indices_far = (distances >= vanderwaals_radius) & (distances < 1.5 * vanderwaals_radius)

        density_data[indices_close] = np.exp(-2.0 * np.square(distances[indices_close]) / np.square(vanderwaals_radius))
        density_data[indices_far] = (
            4.0 / np.square(np.e) / np.square(vanderwaals_radius) * np.square(distances[indices_far])
            - 12.0 / np.square(np.e) / vanderwaals_radius * distances[indices_far]
            + 9.0 / np.square(np.e)
        )

        return density_data

    def map_feature(
        self,
        position: NDArray,
        feature_name: str,
        feature_value: NDArray | float,
        method: MapMethod,
    ) -> None:
        """Maps point feature data at a given position to the grid, using the given method.

        The feature_value should either be a single number or a one-dimensional array.
        """
        # determine whether we're dealing with a single number of multiple numbers:
        index_names_values = []
        if isinstance(feature_value, float):
            index_names_values = [(feature_name, feature_value)]

        elif isinstance(feature_value, int):
            index_names_values = [(feature_name, float(feature_value))]

        else:
            for index, value in enumerate(feature_value):
                index_name = f"{feature_name}_{index:03d}"
                index_names_values.append((index_name, value))

        # map the data to the grid
        for index_name, value in index_names_values:
            if method == MapMethod.GAUSSIAN:
                grid_data = self._get_mapped_feature_gaussian(position, value)

            elif method == MapMethod.FAST_GAUSSIAN:
                grid_data = self._get_mapped_feature_fast_gaussian(position, value)

            elif method == MapMethod.BSP_LINE:
                grid_data = self._get_mapped_feature_bsp_line(position, value)

            elif method == MapMethod.NEAREST_NEIGHBOURS:
                grid_data = self._get_mapped_feature_nearest_neighbour(position, value)

            # set to grid
            self.add_feature_values(index_name, grid_data)

    def to_hdf5(self, hdf5_path: str) -> None:
        """Write the grid data to hdf5, according to deeprank standards."""
        with h5py.File(hdf5_path, "a") as hdf5_file:
            # create a group to hold everything
            grid_group = hdf5_file.require_group(self.id)

            # store grid points
            points_group = grid_group.require_group("grid_points")
            points_group.create_dataset("x", data=self.xs)
            points_group.create_dataset("y", data=self.ys)
            points_group.create_dataset("z", data=self.zs)
            points_group.create_dataset("center", data=self.center)

            # store grid features
            features_group = grid_group.require_group(gridstorage.MAPPED_FEATURES)
            for feature_name, feature_data in self.features.items():
                features_group.create_dataset(
                    feature_name,
                    data=feature_data,
                    compression="lzf",
                    chunks=True,
                )



from torch import nn

regression_losses = (
    nn.L1Loss,
    nn.SmoothL1Loss,
    nn.MSELoss,
    nn.HuberLoss,
)

binary_classification_losses = (
    nn.SoftMarginLoss,
    nn.BCELoss,
    nn.BCEWithLogitsLoss,
)

multi_classification_losses = (
    nn.CrossEntropyLoss,
    nn.NLLLoss,
    nn.PoissonNLLLoss,
    nn.GaussianNLLLoss,
    nn.KLDivLoss,
    nn.MultiLabelMarginLoss,
    nn.MultiLabelSoftMarginLoss,
)

other_losses = (
    nn.HingeEmbeddingLoss,
    nn.CosineEmbeddingLoss,
    nn.MarginRankingLoss,
    nn.TripletMarginLoss,
    nn.CTCLoss,
)

classification_losses = multi_classification_losses + binary_classification_losses

classification_tested = (
    nn.CrossEntropyLoss,
    nn.NLLLoss,
    nn.BCELoss,
    nn.BCEWithLogitsLoss,
)



from __future__ import annotations

from typing import TYPE_CHECKING

from typing_extensions import Self

if TYPE_CHECKING:
    from deeprank2.molstruct.atom import Atom
    from deeprank2.molstruct.residue import Residue
    from deeprank2.utils.pssmdata import PssmRow


class PDBStructure:
    """."""

    def __init__(self, id_: str | None = None):
        """A proitein or protein complex structure.

        A `PDBStructure` can contain one or multiple `Chains`, i.e. separate molecular entities (individual proteins).
        One PDBStructure consists of a number of `Residue`s, each of which is of a particular `AminoAcid` type and in turn consists of a number of `Atom`s.

        Args:
            id_ (str, optional): An unique identifier for this structure, can be the pdb accession code.
                Defaults to None.
        """
        self._id = id_
        self._chains = {}

    def __eq__(self, other: Self) -> bool:
        if isinstance(other, PDBStructure):
            return self._id == other._id
        return NotImplemented

    def __hash__(self) -> hash:
        return hash(self._id)

    def __repr__(self) -> str:
        return self._id

    def has_chain(self, chain_id: str) -> bool:
        return chain_id in self._chains

    def get_chain(self, chain_id: str) -> Chain:
        return self._chains[chain_id]

    def add_chain(self, chain: Chain) -> None:
        if chain.id in self._chains:
            msg = f"Duplicate chain: {chain.id}"
            raise ValueError(msg)
        self._chains[chain.id] = chain

    @property
    def chains(self) -> list[Chain]:
        return list(self._chains.values())

    def get_atoms(self) -> list[Atom]:
        """List all atoms in the structure."""
        atoms = []
        for chain in self._chains.values():
            for residue in chain.residues:
                atoms.extend(residue.atoms)
        return atoms

    @property
    def id(self) -> str:
        return self._id


class Chain:
    """One independent molecular entity of a `PDBStructure`.

    In other words: each `Chain` in a `PDBStructure` is a separate molecule.
    """

    def __init__(self, model: PDBStructure, id_: str | None):
        """One chain of a PDBStructure.

        Args:
        model (:class:`PDBStructure`): The model that this chain is part of.
        id_ (str): The pdb identifier of this chain.
        """
        self._model = model
        self._id = id_
        self._residues = {}
        self._pssm = None  # pssm is per chain

    @property
    def model(self) -> PDBStructure:
        return self._model

    @property
    def pssm(self) -> PssmRow:
        return self._pssm

    @pssm.setter
    def pssm(self, pssm: PssmRow) -> None:
        self._pssm = pssm

    def add_residue(self, residue: Residue) -> None:
        self._residues[(residue.number, residue.insertion_code)] = residue

    def has_residue(self, residue_number: int, insertion_code: str | None = None) -> bool:
        return (residue_number, insertion_code) in self._residues

    def get_residue(self, residue_number: int, insertion_code: str | None = None) -> Residue:
        return self._residues[(residue_number, insertion_code)]

    @property
    def id(self) -> str:
        return self._id

    @property
    def residues(self) -> list[Residue]:
        return list(self._residues.values())

    def get_atoms(self) -> list[Atom]:
        """Shortcut to list all atoms in this chain."""
        atoms = []
        for residue in self.residues:
            atoms.extend(residue.atoms)

        return atoms

    def __eq__(self, other: Self) -> bool:
        if isinstance(other, Chain):
            return self._model == other._model and self._id == other._id
        return NotImplemented

    def __hash__(self) -> hash:
        return hash(self._id)

    def __repr__(self) -> str:
        return f"{self._model} {self._id}"



import torch
import torch.nn
import torch.nn.functional as F
from torch.autograd import Variable

# ruff: noqa: ANN001, ANN201, ANN202

######################################################################
#
# Model automatically generated by modelGenerator
#
######################################################################

# ----------------------------------------------------------------------
# Network Structure
# ----------------------------------------------------------------------
# conv layer   0: conv | input -1  output  4  kernel  2  post relu
# conv layer   1: pool | kernel  2  post None
# conv layer   2: conv | input  4  output  5  kernel  2  post relu
# conv layer   3: pool | kernel  2  post None
# fc   layer   0: fc   | input -1  output  84  post relu
# fc   layer   1: fc   | input  84  output  1  post None
# ----------------------------------------------------------------------


class CnnRegression(torch.nn.Module):  # noqa: D101
    def __init__(self, num_features: int, box_shape: tuple[int]):
        super().__init__()

        self.convlayer_000 = torch.nn.Conv3d(num_features, 4, kernel_size=2)
        self.convlayer_001 = torch.nn.MaxPool3d((2, 2, 2))
        self.convlayer_002 = torch.nn.Conv3d(4, 5, kernel_size=2)
        self.convlayer_003 = torch.nn.MaxPool3d((2, 2, 2))

        size = self._get_conv_output(num_features, box_shape)

        self.fclayer_000 = torch.nn.Linear(size, 84)
        self.fclayer_001 = torch.nn.Linear(84, 1)

    def _get_conv_output(self, num_features: int, shape: tuple[int]):
        num_data_points = 2
        input_ = Variable(torch.rand(num_data_points, num_features, *shape))
        output = self._forward_features(input_)
        return output.data.view(num_data_points, -1).size(1)

    def _forward_features(self, x):
        x = F.relu(self.convlayer_000(x))
        x = self.convlayer_001(x)
        x = F.relu(self.convlayer_002(x))
        x = self.convlayer_003(x)
        return x  # noqa:RET504 (unnecessary-assign)

    def forward(self, data):
        x = self._forward_features(data.x)
        x = x.view(x.size(0), -1)
        x = F.relu(self.fclayer_000(x))
        x = self.fclayer_001(x)
        return x  # noqa:RET504 (unnecessary-assign)


######################################################################
#
# Model automatically generated by modelGenerator
#
######################################################################

# ----------------------------------------------------------------------
# Network Structure
# ----------------------------------------------------------------------
# conv layer   0: conv | input -1  output  4  kernel  2  post relu
# conv layer   1: pool | kernel  2  post None
# conv layer   2: conv | input  4  output  5  kernel  2  post relu
# conv layer   3: pool | kernel  2  post None
# fc   layer   0: fc   | input -1  output  84  post relu
# fc   layer   1: fc   | input  84  output  1  post None
# ----------------------------------------------------------------------


class CnnClassification(torch.nn.Module):  # noqa: D101
    def __init__(self, num_features, box_shape):
        super().__init__()

        self.convlayer_000 = torch.nn.Conv3d(num_features, 4, kernel_size=2)
        self.convlayer_001 = torch.nn.MaxPool3d((2, 2, 2))
        self.convlayer_002 = torch.nn.Conv3d(4, 5, kernel_size=2)
        self.convlayer_003 = torch.nn.MaxPool3d((2, 2, 2))

        size = self._get_conv_output(num_features, box_shape)

        self.fclayer_000 = torch.nn.Linear(size, 84)
        self.fclayer_001 = torch.nn.Linear(84, 2)

    def _get_conv_output(self, num_features, shape):
        inp = Variable(torch.rand(1, num_features, *shape))
        out = self._forward_features(inp)
        return out.data.view(1, -1).size(1)

    def _forward_features(self, x):
        x = F.relu(self.convlayer_000(x))
        x = self.convlayer_001(x)
        x = F.relu(self.convlayer_002(x))
        x = self.convlayer_003(x)
        return x  # noqa:RET504 (unnecessary-assign)

    def forward(self, data):
        x = self._forward_features(data.x)
        x = x.view(x.size(0), -1)
        x = F.relu(self.fclayer_000(x))
        x = self.fclayer_001(x)
        return x  # noqa:RET504 (unnecessary-assign)



from abc import ABC
from typing import Any

from typing_extensions import Self

from deeprank2.molstruct.atom import Atom
from deeprank2.molstruct.residue import Residue


class Pair:
    """A hashable, comparable object for any set of two inputs where order doesn't matter.

    Args:
        item1 (Any object): The pair's first object, must be convertable to string.
        item2 (Any object): The pair's second object, must be convertable to string.
    """

    def __init__(self, item1: Any, item2: Any):  # noqa: ANN401
        self.item1 = item1
        self.item2 = item2

    def __hash__(self) -> hash:
        """The hash should be solely based on the two paired items, not on their order."""
        s1 = str(self.item1)
        s2 = str(self.item2)
        if s1 < s2:
            return hash(s1 + s2)
        return hash(s2 + s1)

    def __eq__(self, other: Self) -> bool:
        """Compare the pairs as sets, so the order doesn't matter."""
        if isinstance(other, Pair):
            return self.item1 == other.item1 and self.item2 == other.item2 or self.item1 == other.item2 and self.item2 == other.item1
        return NotImplemented

    def __iter__(self):
        # Iterate over the two items in the pair.
        return iter([self.item1, self.item2])

    def __repr__(self) -> str:
        return str(self.item1) + str(self.item2)


class Contact(Pair, ABC):
    """Parent class to bind `ResidueContact` and `AtomicContact` objects."""


class ResidueContact(Contact):
    """A contact between two residues from a structure."""

    def __init__(self, residue1: Residue, residue2: Residue):
        self._residue1 = residue1
        self._residue2 = residue2
        super().__init__(residue1, residue2)

    @property
    def residue1(self) -> Residue:
        return self.item1

    @property
    def residue2(self) -> Residue:
        return self.item2


class AtomicContact(Contact):
    """A contact between two atoms from a structure."""

    def __init__(self, atom1: Atom, atom2: Atom):
        self._atom1 = atom1
        self._atom2 = atom2
        super().__init__(atom1, atom2)

    @property
    def atom1(self) -> Atom:
        return self.item1

    @property
    def atom2(self) -> Atom:
        return self.item2



import warnings

import community
import markov_clustering as mc
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import torch
from torch_geometric.data import Batch, Data
from torch_geometric.nn.pool.consecutive import consecutive_cluster
from torch_geometric.nn.pool.pool import pool_batch, pool_edge
from torch_scatter import scatter_max, scatter_mean

# ruff: noqa: ANN001, ANN201


def plot_graph(graph, cluster) -> None:  # noqa:D103
    pos = nx.spring_layout(graph, iterations=200)
    nx.draw(graph, pos, node_color=cluster)
    plt.show()


def get_preloaded_cluster(cluster, batch):  # noqa:D103
    nbatch = torch.max(batch) + 1
    for ib in range(1, nbatch):
        cluster[batch == ib] += torch.max(cluster[batch == ib - 1]) + 1
    return cluster


def community_detection_per_batch(
    edge_index,
    batch,
    num_nodes: int,
    edge_attr=None,
    method: str = "mcl",
):
    """Detects clusters of nodes based on the edge attributes (distances).

    Args:
        edge_index (Tensor): Edge index.
        batch (?): ?
        num_nodes (int): Number of nodes.
        edge_attr (Tensor, optional): Edge attributes. Defaults to None.
        method (str, optional): Method. Defaults to "mcl".

    Raises:
        ValueError: Requires a valid clustering method ('mcl' or 'louvain')

    Returns:
        cluster Tensor
    """
    # make the networkX graph
    g = nx.Graph()
    g.add_nodes_from(range(num_nodes))

    for iedge, (i, j) in enumerate(edge_index.transpose(0, 1).tolist()):
        if edge_attr is None:
            g.add_edge(i, j)
        else:
            g.add_edge(i, j, weight=edge_attr[iedge])

    num_batch = max(batch) + 1
    all_index = range(num_nodes)
    cluster, ncluster = [], 0

    for ib in range(num_batch):
        index = torch.tensor(all_index)[batch == ib].tolist()
        subg = g.subgraph(index)

        # detect the communities using Louvain method
        if method == "louvain":
            c = community.best_partition(subg)
            cluster += [v + ncluster for k, v in c.items()]
            ncluster = max(cluster)

        # detect communities using MCL
        elif method == "mcl":
            matrix = nx.to_scipy_sparse_array(subg)
            # run MCL with default parameters
            result = mc.run_mcl(matrix.toarray())
            mc_clust = mc.get_clusters(result)  # get clusters

            index = np.zeros(subg.number_of_nodes()).astype("int")
            for ic, c in enumerate(mc_clust):
                index[list(c)] = ic + ncluster
            cluster += index.tolist()
            ncluster = max(cluster)

        else:
            msg = f"Clustering method {method} not supported"
            raise ValueError(msg)
    device = edge_index.device
    return torch.tensor(cluster).to(device)


def community_detection(
    edge_index,
    num_nodes: int,
    edge_attr=None,
    method: str = "mcl",
):
    """Detects clusters of nodes based on the edge attributes (distances).

    Args:
        edge_index (Tensor): Edge index.
        num_nodes (int): Number of nodes.
        edge_attr (Tensor, optional): Edge attributes. Defaults to None.
        method (str, optional): Method. Defaults to "mcl".

    Raises:
        ValueError: Requires a valid clustering method ('mcl' or 'louvain')

    Returns:
        cluster Tensor

    Examples:
        >>> import torch
        >>> from torch_geometric.data import Data, Batch
        >>> edge_index = torch.tensor([[0, 1, 1, 2, 3, 4, 4, 5],
        >>>                            [1, 0, 2, 1, 4, 3, 5, 4]], dtype=torch.long)
        >>> x = torch.tensor([[0], [1], [2], [3], [4], [5]],
        >>>                  dtype=torch.float)
        >>> data = Data(x=x, edge_index=edge_index)
        >>> data.pos = torch.tensor(np.random.rand(data.num_nodes, 3))
        >>> c = community_detection(data.edge_index, data.num_nodes)
    """
    # make the networkX graph
    g = nx.Graph()
    g.add_nodes_from(range(num_nodes))

    for iedge, (i, j) in enumerate(edge_index.transpose(0, 1).tolist()):
        if edge_attr is None:
            g.add_edge(i, j)
        else:
            g.add_edge(i, j, weight=edge_attr[iedge])

    # get the device
    device = edge_index.device

    # detect the communities using Louvain detection
    if method == "louvain":
        cluster = community.best_partition(g)
        return torch.tensor([v for k, v in cluster.items()]).to(device)

    # detect the communities using MCL detection
    if method == "mcl":
        matrix = nx.to_scipy_sparse_array(g)

        # run MCL with default parameters

        result = mc.run_mcl(matrix.toarray())

        clusters = mc.get_clusters(result)  # get clusters

        index = np.zeros(num_nodes).astype("int")
        for ic, c in enumerate(clusters):
            index[list(c)] = ic

        return torch.tensor(index).to(device)

    msg = f"Clustering method {method} not supported"
    raise ValueError(msg)


def community_pooling(cluster, data):
    """Pools features and edges of all cluster members.

    All cluster members are pooled into a single node that is assigned:
    - the max cluster value for each feature
    - the average cluster nodes position

    Args:
        cluster: Clusters.
        data: Features tensor.

    Returns:
        pooled features tensor

    Examples:
        >>> import torch
        >>> from torch_geometric.data import Data, Batch
        >>> edge_index = torch.tensor([[0, 1, 1, 2, 3, 4, 4, 5],
        >>>                            [1, 0, 2, 1, 4, 3, 5, 4]], dtype=torch.long)
        >>> x = torch.tensor([[0], [1], [2], [3], [4], [5]],
        >>>                  dtype=torch.float)
        >>> data = Data(x=x, edge_index=edge_index)
        >>> data.pos = torch.tensor(np.random.rand(data.num_nodes, 3))
        >>> c = community_detection(data.edge_index, data.num_nodes)
        >>> batch = Batch().from_data_list([data, data])
        >>> cluster = community_detection(batch.edge_index, batch.num_nodes)
        >>> new_batch = community_pooling(cluster, batch)
    """
    # determine what the batches has as attributes
    has_internal_edges = hasattr(data, "internal_edge_index")
    has_pos2d = hasattr(data, "pos2d")
    has_pos = hasattr(data, "pos")
    has_cluster = hasattr(data, "cluster0")

    if has_internal_edges:
        warnings.warn(
            """Internal edges are not supported anymore. Please prepare the hdf5 file with a more up to date version of this software.""",
            DeprecationWarning,
        )

    cluster, perm = consecutive_cluster(cluster)
    cluster = cluster.to(data.x.device)

    # pool the node infos
    x, _ = scatter_max(data.x, cluster, dim=0)

    # pool the edges
    edge_index, edge_attr = pool_edge(cluster, data.edge_index, data.edge_attr)

    # pool the pos
    if has_pos:
        pos = scatter_mean(data.pos, cluster, dim=0)
    if has_pos2d:
        pos2d = scatter_mean(data.pos2d, cluster, dim=0)

    if has_cluster:
        c0, c1 = data.cluster0, data.cluster1

    # pool batch
    if hasattr(data, "batch"):
        batch = None if data.batch is None else pool_batch(perm, data.batch)
        data = Batch(batch=batch, x=x, edge_index=edge_index, edge_attr=edge_attr, pos=pos)

        if has_cluster:
            data.cluster0 = c0
            data.cluster1 = c1

    else:
        data = Data(x=x, edge_index=edge_index, edge_attr=edge_attr, pos=pos)

        if has_pos2d:
            data.pos2d = pos2d

        if has_cluster:
            data.cluster0 = c0
            data.cluster1 = c1

    return data



# list of default targets and tasks in hdf5 files

## main group
VALUES = "target_values"

## classification tasks
BINARY = "binary"
CAPRI = "capri_class"

## regression tasks
IRMSD = "irmsd"
LRMSD = "lrmsd"
FNAT = "fnat"
DOCKQ = "dockq"

## task names
REGRESS = "regress"
CLASSIF = "classif"



## main group
NODE = "node_features"

## metafeatures
NAME = "_name"
CHAINID = "_chain_id"  # str; former FEATURENAME_CHAIN (was not assigned, but supposedly numeric, now a str)
POSITION = "_position"  # list[3xfloat]; former FEATURENAME_POSITION

## atom core features
ATOMTYPE = "atom_type"
ATOMCHARGE = "atom_charge"
PDBOCCUPANCY = "pdb_occupancy"

## residue core features
RESTYPE = "res_type"  # AminoAcid object; former FEATURENAME_AMINOACID
RESCHARGE = "res_charge"  # float(<0); former FEATURENAME_CHARGE (was not assigned)
POLARITY = "polarity"  #  Polarity object; former FEATURENAME_POLARITY
RESSIZE = "res_size"  # int; former FEATURENAME_SIZE
RESMASS = "res_mass"
RESPI = "res_pI"
HBDONORS = "hb_donors"  # int; former FEATURENAME_HYDROGENBONDDONORS
HBACCEPTORS = "hb_acceptors"  # int; former FEATURENAME_HYDROGENBONDACCEPTORS

## variant residue features
VARIANTRES = "variant_res"  # AminoAcid object; former FEATURENAME_VARIANTAMINOACID
DIFFCHARGE = "diff_charge"  # float
DIFFSIZE = "diff_size"  # int; former FEATURENAME_SIZEDIFFERENCE
DIFFMASS = "diff_mass"
DIFFPI = "diff_pI"
DIFFPOLARITY = "diff_polarity"  # [type?]; former FEATURENAME_POLARITYDIFFERENCE
DIFFHBDONORS = "diff_hb_donors"  # int; former FEATURENAME_HYDROGENBONDDONORSDIFFERENCE
DIFFHBACCEPTORS = "diff_hb_acceptors"  # int; former FEATURENAME_HYDROGENBONDACCEPTORSDIFFERENCE

## conservation features
PSSM = "pssm"  # list[20xint]; former FEATURENAME_PSSM
INFOCONTENT = "info_content"  # float; former FEATURENAME_INFORMATIONCONTENT
CONSERVATION = "conservation"  # int; former FEATURENAME_PSSMWILDTYPE
DIFFCONSERVATION = "diff_conservation"  # int; former FEATURENAME_PSSMDIFFERENCE & FEATURENAME_CONSERVATIONDIFFERENCE

## protein context features
RESDEPTH = "res_depth"  # float; former FEATURENAME_RESIDUEDEPTH
HSE = "hse"  # list[3xfloat]; former FEATURENAME_HALFSPHEREEXPOSURE
SASA = "sasa"  # float; former FEATURENAME_SASA
BSA = "bsa"  # float; former FEATURENAME_BURIEDSURFACEAREA
SECSTRUCT = "sec_struct"  # secondary structure

## inter-residue contacts (IRCs)
IRC_NONNON = "irc_nonpolar_nonpolar"  # int
IRC_NONPOL = "irc_nonpolar_polar"  # int
IRC_NONNEG = "irc_nonpolar_negative"  # int
IRC_NONPOS = "irc_nonpolar_positive"  # int
IRC_POLPOL = "irc_polar_polar"  # int
IRC_POLNEG = "irc_polar_negative"  # int
IRC_POLPOS = "irc_polar_positive"  # int
IRC_NEGNEG = "irc_negative_negative"  # int
IRC_NEGPOS = "irc_negative_positive"  # int
IRC_POSPOS = "irc_positive_positive"  # int
IRCTOTAL = "irc_total"  # int

IRC_FEATURES = [
    IRC_NONNON,
    IRC_NONPOL,
    IRC_NONNEG,
    IRC_NONPOS,
    IRC_POLPOL,
    IRC_POLNEG,
    IRC_POLPOS,
    IRC_NEGNEG,
    IRC_POSPOS,
    IRC_NEGPOS,
    IRCTOTAL,
]



# Example of network that doesn't require clusters.

import torch
from torch.nn import Linear, Module, ReLU, Sequential
from torch_scatter import scatter_mean, scatter_sum

# ruff: noqa: ANN001, ANN201


class NaiveConvolutionalLayer(Module):  # noqa: D101
    def __init__(self, count_node_features, count_edge_features):
        super().__init__()
        message_size = 32
        edge_input_size = 2 * count_node_features + count_edge_features
        self._edge_mlp = Sequential(Linear(edge_input_size, message_size), ReLU())
        node_input_size = count_node_features + message_size
        self._node_mlp = Sequential(Linear(node_input_size, count_node_features), ReLU())

    def forward(self, node_features, edge_node_indices, edge_features):
        # generate messages over edges
        node0_indices, node1_indices = edge_node_indices
        node0_features = node_features[node0_indices]
        node1_features = node_features[node1_indices]
        message_input = torch.cat([node0_features, node1_features, edge_features], dim=1)
        messages_per_neighbour = self._edge_mlp(message_input)
        # aggregate messages
        out = torch.zeros(node_features.shape[0], messages_per_neighbour.shape[1]).to(node_features.device)
        message_sums_per_node = scatter_sum(messages_per_neighbour, node0_indices, dim=0, out=out)
        # update nodes
        node_input = torch.cat([node_features, message_sums_per_node], dim=1)
        return self._node_mlp(node_input)


class NaiveNetwork(Module):  # noqa: D101
    def __init__(self, input_shape: int, output_shape: int, input_shape_edge: int):
        """NaiveNetwork.

        Args:
            input_shape (int): Number of node input features.
            output_shape (int): Number of output value per graph.
            input_shape_edge (int): Number of edge input features.
        """
        super().__init__()
        self._external1 = NaiveConvolutionalLayer(input_shape, input_shape_edge)
        self._external2 = NaiveConvolutionalLayer(input_shape, input_shape_edge)
        hidden_size = 128
        self._graph_mlp = Sequential(Linear(input_shape, hidden_size), ReLU(), Linear(hidden_size, output_shape))

    def forward(self, data):
        external_updated1_node_features = self._external1(data.x, data.edge_index, data.edge_attr)
        external_updated2_node_features = self._external2(external_updated1_node_features, data.edge_index, data.edge_attr)
        means_per_graph_external = scatter_mean(external_updated2_node_features, data.batch, dim=0)
        graph_input = means_per_graph_external
        z = self._graph_mlp(graph_input)
        return z  # noqa:RET504 (unnecessary-assign)



import logging
import warnings

import numpy as np
from numpy.typing import NDArray
from scipy.spatial import distance_matrix

from deeprank2.domain import edgestorage as Efeat
from deeprank2.molstruct.atom import Atom
from deeprank2.molstruct.pair import AtomicContact, ResidueContact
from deeprank2.molstruct.residue import SingleResidueVariant
from deeprank2.utils.graph import Graph
from deeprank2.utils.parsing import atomic_forcefield

_log = logging.getLogger(__name__)

# for cutoff distances, see: https://github.com/DeepRank/deeprank2/issues/357#issuecomment-1461813723
covalent_cutoff = 2.1
cutoff_13 = 3.6
cutoff_14 = 4.2
EPSILON0 = 1.0
COULOMB_CONSTANT = 332.0636


def _get_nonbonded_energy(
    atoms: list[Atom],
    distances: NDArray[np.float64],
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Calculates all pairwise electrostatic (Coulomb) and Van der Waals (Lennard Jones) potential energies between all atoms in the structure.

    Warning: there's no distance cutoff here. The radius of influence is assumed to infinite.
    However, the potential tends to 0 at large distance.

    Args:
        atoms (list[Atom]): list of all atoms in the structure
        distances (NDArray[np.float64]): matrix of pairwise distances between all atoms in the structure
            in the format that is the output of scipy.spatial's distance_matrix (i.e. a diagonally symmetric matrix)

    Returns:
        Tuple [NDArray[np.float64], NDArray[np.float64]]: matrices in same format as `distances` containing
            all pairwise electrostatic potential energies and all pairwise Van der Waals potential energies
    """
    # ELECTROSTATIC POTENTIAL
    charges = [atomic_forcefield.get_charge(atom) for atom in atoms]
    E_elec = np.expand_dims(charges, axis=1) * np.expand_dims(charges, axis=0) * COULOMB_CONSTANT / (EPSILON0 * distances)

    # VAN DER WAALS POTENTIAL
    # calculate main vdw energies
    sigmas = [atomic_forcefield.get_vanderwaals_parameters(atom).sigma_main for atom in atoms]
    epsilons = [atomic_forcefield.get_vanderwaals_parameters(atom).epsilon_main for atom in atoms]
    mean_sigmas = 0.5 * np.add.outer(sigmas, sigmas)
    geomean_eps = np.sqrt(np.multiply.outer(epsilons, epsilons))  # sqrt(eps1*eps2)
    E_vdw = 4.0 * geomean_eps * ((mean_sigmas / distances) ** 12 - (mean_sigmas / distances) ** 6)

    # calculate vdw energies for 1-4 pairs
    sigmas = [atomic_forcefield.get_vanderwaals_parameters(atom).sigma_14 for atom in atoms]
    epsilons = [atomic_forcefield.get_vanderwaals_parameters(atom).epsilon_14 for atom in atoms]
    mean_sigmas = 0.5 * np.add.outer(sigmas, sigmas)
    geomean_eps = np.sqrt(np.multiply.outer(epsilons, epsilons))  # sqrt(eps1*eps2)
    E_vdw_14pairs = 4.0 * geomean_eps * ((mean_sigmas / distances) ** 12 - (mean_sigmas / distances) ** 6)

    # Fix energies for close contacts on same chain
    chains = [atom.residue.chain.id for atom in atoms]
    chain_matrix = [[chain_1 == chain_2 for chain_2 in chains] for chain_1 in chains]
    pair_14 = np.logical_and(distances < cutoff_14, chain_matrix)
    pair_13 = np.logical_and(distances < cutoff_13, chain_matrix)

    E_vdw[pair_14] = E_vdw_14pairs[pair_14]
    E_vdw[pair_13] = 0
    E_elec[pair_13] = 0

    return E_elec, E_vdw


def add_features(  # noqa:D103
    pdb_path: str,  # noqa: ARG001
    graph: Graph,
    single_amino_acid_variant: SingleResidueVariant | None = None,  # noqa: ARG001
) -> None:
    # assign each atoms (from all edges) a unique index
    all_atoms = set()
    if isinstance(graph.edges[0].id, AtomicContact):
        for edge in graph.edges:
            contact = edge.id
            all_atoms.add(contact.atom1)
            all_atoms.add(contact.atom2)
    elif isinstance(graph.edges[0].id, ResidueContact):
        for edge in graph.edges:
            contact = edge.id
            for atom in contact.residue1.atoms + contact.residue2.atoms:
                all_atoms.add(atom)
    else:
        msg = f"Unexpected edge type: {type(graph.edges[0].id)}"
        raise TypeError(msg)

    all_atoms = list(all_atoms)
    atom_dict = {atom: i for i, atom in enumerate(all_atoms)}

    # make pairwise calculations between all atoms in the set
    with warnings.catch_warnings(record=RuntimeWarning):
        warnings.simplefilter("ignore")
        positions = [atom.position for atom in all_atoms]
        interatomic_distances = distance_matrix(positions, positions)
        (
            interatomic_electrostatic_energy,
            interatomic_vanderwaals_energy,
        ) = _get_nonbonded_energy(all_atoms, interatomic_distances)

    # assign features
    for edge in graph.edges:
        contact = edge.id

        if isinstance(contact, AtomicContact):
            ## find the indices
            atom1_index = atom_dict[contact.atom1]
            atom2_index = atom_dict[contact.atom2]
            ## set features
            edge.features[Efeat.SAMERES] = float(contact.atom1.residue == contact.atom2.residue)
            edge.features[Efeat.SAMECHAIN] = float(contact.atom1.residue.chain == contact.atom1.residue.chain)
            edge.features[Efeat.DISTANCE] = interatomic_distances[atom1_index, atom2_index]
            edge.features[Efeat.ELEC] = interatomic_electrostatic_energy[atom1_index, atom2_index]
            edge.features[Efeat.VDW] = interatomic_vanderwaals_energy[atom1_index, atom2_index]

        elif isinstance(contact, ResidueContact):
            ## find the indices
            atom1_indices = [atom_dict[atom] for atom in contact.residue1.atoms]
            atom2_indices = [atom_dict[atom] for atom in contact.residue2.atoms]
            ## set features
            edge.features[Efeat.SAMECHAIN] = float(contact.residue1.chain == contact.residue2.chain)
            edge.features[Efeat.DISTANCE] = np.min([[interatomic_distances[a1, a2] for a1 in atom1_indices] for a2 in atom2_indices])
            edge.features[Efeat.ELEC] = np.sum([[interatomic_electrostatic_energy[a1, a2] for a1 in atom1_indices] for a2 in atom2_indices])
            edge.features[Efeat.VDW] = np.sum([[interatomic_vanderwaals_energy[a1, a2] for a1 in atom1_indices] for a2 in atom2_indices])

        # Calculate irrespective of node type
        edge.features[Efeat.COVALENT] = float(edge.features[Efeat.DISTANCE] < covalent_cutoff and edge.features[Efeat.SAMECHAIN])



import logging
from itertools import combinations_with_replacement as combinations

import pdb2sql

from deeprank2.domain import nodestorage as Nfeat
from deeprank2.domain.aminoacidlist import amino_acids_by_code
from deeprank2.molstruct.aminoacid import Polarity
from deeprank2.molstruct.atom import Atom
from deeprank2.molstruct.residue import Residue, SingleResidueVariant
from deeprank2.utils.graph import Graph

_log = logging.getLogger(__name__)
SAFE_MIN_CONTACTS = 5


def _id_from_residue(residue: tuple[str, int, str]) -> str:
    """Create and id from pdb2sql rendered residues that is similar to the id of residue nodes.

    Args:
        residue (tuple): Input residue as rendered by pdb2sql: ( str(<chain>), int(<residue_number>), str(<three_letter_code> )
            For example: ('A', 27, 'GLU').

    Returns:
        str: Output id in form of '<chain><residue_number>'. For example: 'A27'.
    """
    return residue[0] + str(residue[1])


class _ContactDensity:
    """Internal class that holds contact density information for a given residue."""

    def __init__(self, residue: tuple[str, int, str], polarity: Polarity):
        self.res = residue
        self.polarity = polarity
        self.id = _id_from_residue(self.res)
        self.densities = {pol: 0 for pol in Polarity}
        self.densities["total"] = 0
        self.connections = {pol: [] for pol in Polarity}
        self.connections["all"] = []


def get_IRCs(pdb_path: str, chains: list[str], cutoff: float = 5.5) -> dict[str, _ContactDensity]:
    """Get all close contact residues from the opposite chain.

    Args:
        pdb_path (str): Path to pdb file to read molecular information from.
        chains (Sequence[str]): list (or list-like object) containing strings of the chains to be considered.
        cutoff (float, optional): Cutoff distance (in Ångström) to be considered a close contact. Defaults to 10.

    Returns:
        Dict[str, _ContactDensity]:
            keys: ids of residues in form returned by id_from_residue.
            items: _ContactDensity objects, containing all contact density information for the residue.
    """
    residue_contacts: dict[str, _ContactDensity] = {}

    sql = pdb2sql.interface(pdb_path)
    pdb2sql_contacts = sql.get_contact_residues(
        cutoff=cutoff,
        chain1=chains[0],
        chain2=chains[1],
        return_contact_pairs=True,
    )

    for chain1_res, chain2_residues in pdb2sql_contacts.items():
        aa1_code = chain1_res[2]
        try:
            aa1 = amino_acids_by_code[aa1_code]
        except IndexError:
            continue  # skip keys that are not an amino acid

        # add chain1_res to residue_contact dict
        contact1_id = _id_from_residue(chain1_res)
        residue_contacts[contact1_id] = _ContactDensity(chain1_res, aa1.polarity)

        for chain2_res in chain2_residues:
            aa2_code = chain2_res[2]
            try:
                aa2 = amino_acids_by_code[aa2_code]
            except IndexError:
                continue  # skip keys that are not an amino acid

            # populate densities and connections for chain1_res
            residue_contacts[contact1_id].densities["total"] += 1
            residue_contacts[contact1_id].densities[aa2.polarity] += 1
            residue_contacts[contact1_id].connections["all"].append(chain2_res)
            residue_contacts[contact1_id].connections[aa2.polarity].append(chain2_res)

            # add chain2_res to residue_contact dict if it doesn't exist yet
            contact2_id = _id_from_residue(chain2_res)
            if contact2_id not in residue_contacts:
                residue_contacts[contact2_id] = _ContactDensity(chain2_res, aa2.polarity)

            # populate densities and connections for chain2_res
            residue_contacts[contact2_id].densities["total"] += 1
            residue_contacts[contact2_id].densities[aa1.polarity] += 1
            residue_contacts[contact2_id].connections["all"].append(chain1_res)
            residue_contacts[contact2_id].connections[aa1.polarity].append(chain1_res)

    return residue_contacts


def add_features(  # noqa: C901, D103
    pdb_path: str,
    graph: Graph,
    single_amino_acid_variant: SingleResidueVariant | None = None,
) -> None:
    if not single_amino_acid_variant:  # VariantQueries do not use this feature
        polarity_pairs = list(combinations(Polarity, 2))
        polarity_pair_string = [f"irc_{x[0].name.lower()}_{x[1].name.lower()}" for x in polarity_pairs]

        total_contacts = 0
        residue_contacts = get_IRCs(pdb_path, graph.get_all_chains())

        for node in graph.nodes:
            if isinstance(node.id, Residue):
                residue = node.id
            elif isinstance(node.id, Atom):
                atom = node.id
                residue = atom.residue
            else:
                msg = f"Unexpected node type: {type(node.id)}"
                raise TypeError(msg)

            contact_id = residue.chain.id + residue.number_string  # reformat id to be in line with residue_contacts keys

            # initialize all IRC features to 0
            for IRC_type in Nfeat.IRC_FEATURES:
                node.features[IRC_type] = 0

            # load correct values to IRC features
            try:
                node.features[Nfeat.IRCTOTAL] = residue_contacts[contact_id].densities["total"]
                for i, pair in enumerate(polarity_pairs):
                    if residue_contacts[contact_id].polarity == pair[0]:
                        node.features[polarity_pair_string[i]] = residue_contacts[contact_id].densities[pair[1]]
                    elif residue_contacts[contact_id].polarity == pair[1]:
                        node.features[polarity_pair_string[i]] = residue_contacts[contact_id].densities[pair[0]]
                total_contacts += 1
            except KeyError:  # node has no contact residues and all counts remain 0
                pass

        if total_contacts < SAFE_MIN_CONTACTS:
            _log.warning(f"Few ({total_contacts}) contacts detected for {pdb_path}.")



from collections.abc import Callable


class EarlyStopping:
    """Terminate training upon trigger.

    Triggered if validation loss doesn't improve after a given patience or if a maximum gap between validation and training loss is reached.

    Args:
        patience (int, optional): How long to wait after last time validation loss improved.
            Defaults to 10.
        delta (float, optional): Minimum change required to reset the early stopping counter.
            Defaults to 0.
        maxgap (float, optional): Maximum difference between between training and validation loss.
            Defaults to None.
        min_epoch (float, optional): Minimum epoch to be reached before looking at maxgap.
            Defaults to 10.
        verbose (bool, optional): If True, prints a message for each validation loss improvement.
            Defaults to True.
        trace_func (Callable, optional): Function used for recording EarlyStopping status.
            Defaults to print.
    """

    def __init__(
        self,
        patience: int = 10,
        delta: float = 0,
        maxgap: float | None = None,
        min_epoch: int = 10,
        verbose: bool = True,
        trace_func: Callable = print,
    ):
        self.patience = patience
        self.delta = delta
        self.maxgap = maxgap
        self.min_epoch = min_epoch
        self.verbose = verbose
        self.trace_func = trace_func

        self.early_stop = False
        self.counter = 0
        self.best_score = None
        self.val_loss_min = None

    def __call__(  # noqa: C901
        self,
        epoch: int,
        val_loss: float,
        train_loss: float | None = None,
    ):
        score = -val_loss

        # initialize
        if self.best_score is None:
            self.best_score = score
            self.val_loss_min = val_loss

        # check patience
        elif score < self.best_score + self.delta:
            self.counter += 1
            if self.verbose:
                extra_trace = ""
                if self.delta:
                    extra_trace = f"more than {self.delta} "
                self.trace_func(
                    f"Validation loss did not decrease {extra_trace}({self.val_loss_min:.6f} --> {val_loss:.6f}). "
                    f"EarlyStopping counter: {self.counter} out of {self.patience}",
                )
            if self.counter >= self.patience:
                self.trace_func(f"EarlyStopping activated at epoch # {epoch} because patience of {self.patience} has been reached.")
                self.early_stop = True
        else:
            if self.verbose:
                self.trace_func(f"Validation loss decreased ({self.val_loss_min:.6f} --> {val_loss:.6f}).")
            self.best_score = score
            self.counter = 0

        if score >= self.best_score:
            self.best_score = score
            self.val_loss_min = val_loss

        # check maxgap
        if self.maxgap and epoch > self.min_epoch:
            if train_loss is None:
                msg = "Cannot compute gap because no train_loss is provided to EarlyStopping."
                raise ValueError(msg)
            gap = val_loss - train_loss
            if gap > self.maxgap:
                self.trace_func(
                    f"EarlyStopping activated at epoch # {epoch} due to overfitting. "
                    f"The difference between validation and training loss of {gap} exceeds the maximum allowed ({self.maxgap})",
                )
                self.early_stop = True


# This module is modified from https://github.com/Bjarten/early-stopping-pytorch, under the following license:


# MIT License

# Copyright (c) 2018 Bjarte Mehus Sunde

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import torch
import torch.nn.functional as F
from torch import nn
from torch.nn import Parameter
from torch_geometric.nn import max_pool_x
from torch_geometric.nn.inits import uniform
from torch_scatter import scatter_mean

from deeprank2.utils.community_pooling import community_pooling, get_preloaded_cluster

# ruff: noqa: ANN001, ANN201


class FoutLayer(torch.nn.Module):
    """FoutLayer.

    This layer is described by eq. (1) of
    Protein Interface Predition using Graph Convolutional Network
    by Alex Fout et al. NIPS 2018.

    Args:
        in_channels (int): Size of each input sample.
        out_channels (int): Size of each output sample.
        bias (bool, optional): If set to :obj:`False`, the layer will not learn
            an additive bias. Defaults to True.
    """

    def __init__(self, in_channels: int, out_channels: int, bias: bool = True):
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels

        # Wc and Wn are the center and neighbor weight matrix
        self.wc = Parameter(torch.Tensor(in_channels, out_channels))
        self.wn = Parameter(torch.Tensor(in_channels, out_channels))

        if bias:
            self.bias = Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        size = self.in_channels
        uniform(size, self.wc)
        uniform(size, self.wn)
        uniform(size, self.bias)

    def forward(self, x, edge_index):
        num_node = len(x)
        alpha = torch.mm(x, self.wc)
        beta = torch.mm(x, self.wn)

        # gamma_i = 1/Ni Sum_j x_j * Wn
        # there might be a better way than looping over the nodes
        gamma = torch.zeros(num_node, self.out_channels).to(alpha.device)
        for n in range(num_node):
            index = edge_index[:, edge_index[0, :] == n][1, :]
            gamma[n, :] = torch.mean(beta[index, :], dim=0)

        alpha = alpha + gamma

        # add the bias
        if self.bias is not None:
            alpha = alpha + self.bias

        return alpha

    def __repr__(self):
        return f"{self.__class__.__name__}({self.in_channels}, {self.out_channels})"


class FoutNet(torch.nn.Module):  # noqa: D101
    def __init__(
        self,
        input_shape,
        output_shape=1,
        input_shape_edge=None,  # noqa: ARG002
    ):
        super().__init__()

        self.conv1 = FoutLayer(input_shape, 16)
        self.conv2 = FoutLayer(16, 32)

        self.fc1 = torch.nn.Linear(32, 64)
        self.fc2 = torch.nn.Linear(64, output_shape)

        self.clustering = "mcl"

    def forward(self, data):
        act = nn.Tanhshrink()
        act = F.relu

        # first conv block
        data.x = act(self.conv1(data.x, data.edge_index))
        cluster = get_preloaded_cluster(data.cluster0, data.batch)
        data = community_pooling(cluster, data)

        # second conv block
        data.x = act(self.conv2(data.x, data.edge_index))
        cluster = get_preloaded_cluster(data.cluster1, data.batch)
        x, batch = max_pool_x(cluster, data.x, data.batch)

        # FC
        x = scatter_mean(x, batch, dim=0)
        x = act(self.fc1(x))
        x = self.fc2(x)

        return x  # noqa:RET504 (unnecessary-assign)
